/*
 * mp3plot - Bitrate analysis tool
 *
 * Copyright (C) 2007, 2009 Toni Corvera
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

// $Id: common.h 1150 2009-04-14 02:21:18Z  $

#ifndef MP3PLOT_MP3COMMON_H
#define MP3PLOT_MP3COMMON_H

#include <string>
#include <list>
#include <stdint.h>

#include "mp3_config.h" // Defines important macros and has common conditional includes
#include "errors.h"

namespace mp3plot {

    namespace mp3 {
        // Forward declaration
        class frame;
    }

    /*!
     * \brief Shorthand for a common cast: Consider the input pointer
     *        a char pointer.
     * \note Used to call read functions in a more compact way
     * \param[in,out] ptr Pointer to reinterpret
     * \return \em a reinterpreted as a <code>char *</code>
     */
    inline char * const CHARR(void * const ptr) {
        return reinterpret_cast<char * const>(ptr);
    }

    /*!
     * \brief Type for bitrate in kbps
     * \remarks The maximum non-free-form bitrate is 384, free format bitrates
     *          are theoretically unlimited although most decoders impose a
     *          real-world limit (AFAIK well below 1000kbps).
     *          Variables of this type are guaranteed to be able to hold any
     *          (real world) bitrate AND the special \em bad-bitrate value.
     * \remarks \em The bad-bitrate is guaranteed to be a negative (and thus
     *          invalid) bitrate, hence this type must also be signed.
     * \see standard_bitrates::bad_bitrate()
     */ // short is guaranteed to be >= 16 bits in C++
    typedef signed short bitrate_t;

    /*!
     * \brief Type for numbers of frames
     * \remarks The number of frames in a standard file is stored in a 32 bits
     *          unsigned value, this type is guaranteed to be able to hold
     *          such a value (hence, at least 32 bits, and able to hold the
     *          max 32 bits unsigned integer.
     * \note Note that a 64 bits unsigned integer would also fit this contract).
     */ // long is guaranteed to be >= 32 bits in C++
    typedef unsigned long numframes_t;

    /*!
     * \brief MPEG Version
     * \note The values match the bit in the header
     */
    enum mpeg_version_t {
        MPEG2_5 = 0x00,     //!< MPEG 2.5 (unofficial)
        MPEG_VERSION_RESERVED = 0x01,   //<! Reserved value
        MPEG2 = 0x02,       //!< MPEG-2
        MPEG1 = 0x03        //!< MPEG-1
    };

    /*!
     * \brief MPEG Layer
     * \note The values match the bit in the header
     */
    enum mpeg_layer_t {
        LAYER_RESERVED = 0,   //!< Reserved value
        LAYER_III,            //!< Layer III (MP3)
        LAYER_II,             //!< Layer II (MP2)
        LAYER_I               //!< Layer I (MP1)
    };

    /*!
     * \brief Channel mode
     * \note The values match the bit in the header
     */
    enum channel_mode_t {
        STEREO = 0,         //!< Stereo
        JOINT_STEREO = 1,   //!< Joint-Stereo
        DUAL = 2,           //!< Dual-Mono (two mono channels)
        DUAL_MONO = DUAL,   //!< Alias for Dual-Mono (two mono channels)
        MONO = 3,           //!< Mono (one channel)

        CHANNEL_MODE_UNDEFINED = 5  //!< Not in the spec, used internally
    };

    /*!
     * \brief Possible flags in a VBR header.
     * These flags indicate if some details of the file are included
     * in the header.
     */
    enum vbr_flags_t {
        FL_NUMFRAMES = 0x01, //!< Number of frames is stored
        FL_FILESIZE  = 0x02, //!< Filesize (in bytes) is stored
        FL_TOC       = 0x04, //!< TOC is stored
        FL_VBR_SCALE = 0x08  //!< VBR scale is stored
    };

    /*!
     * \brief Variants of mp3 file structures.
     * This refers to the file structure not to the encoding used (i.e. if
     * the file start is an ID3v2 tag, an MP3 frame or none.
     * \note Technically a file could still be a valid MP3 even being "UNKNOWN" (i.e.
     * if it's a broken stream).
     */
    enum mp3_filetype_t {
        FT_MP3,     //!< The file starts with an MP3 Frame
        FT_ID3V2,   //!< The file starts with an ID3v2
        UNKNOWN     //!< The file starts with something else
    };

    /*!
     * \brief Collection of frames.
     * No guarantees are made on the container type.
     */
    // Details on the requirements:
    //  * There's no need for random access (except for the first frame)
    //  * Access is sequential
    //  * While reading the file insertion will be one at a time, an unknown
    //    number of times.
    typedef std::list<mp3::frame> frame_collection;

} // namespace mp3plot

#endif // MP3PLOT_MP3COMMON_H
// vim:set ts=4 et ai:

