/*
 * mp3plot - Bitrate analysis tool
 *
 * Copyright (C) 2007, 2009 Toni Corvera
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

// $Id: bitrate.cc 1342 2009-06-21 23:41:45Z toni $

#include "bitrate.h"

#include <iostream>
#include <bitset>
#include <cassert>

#include <map>

namespace {
    using namespace mp3plot;
    using namespace std;

    typedef uint8_t bitrate_index_t;

    typedef pair<mpeg_version_t, mpeg_layer_t> bitrate_map_key;
    typedef vector<bitrate_t> bitrate_list;
    typedef map<bitrate_map_key, bitrate_list> bitrate_map_t;

#define BAD_BITRATE mp3plot::mp3::bitrate_translator::bad_bitrate()

    static const bitrate_t v1l1_brs[] = {
        0, 32, 64, 96, 128, 160, 192, 224, 256, 288, 320, 352, 384, 416, 448, BAD_BITRATE
    };
    static const bitrate_t v1l2_brs[] = {
        0, 32, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, 384, BAD_BITRATE
    };
    static const bitrate_t v1l3_brs[] = {
        0, 32, 40, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, BAD_BITRATE
    };
    static const bitrate_t v2l1_brs[] = {
        0, 32, 48, 56, 64, 80, 96, 112, 128, 144, 160, 176, 192, 224, 256, BAD_BITRATE
    };
    // MPEG2 Layer II and Layer III use the same table
    static const bitrate_t v2l2_brs[] = {
        0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160, BAD_BITRATE
    };

    inline bitrate_t get_mpeg1_layer1(uint8_t header_index) {
        return v1l1_brs[header_index];
    }

    inline bitrate_t get_mpeg1_layer2(uint8_t header_index) {
        return v1l2_brs[header_index];
    }

    inline bitrate_t get_mpeg1_layer3(uint8_t header_index) {
        return v1l3_brs[header_index];
    }

    inline bitrate_t get_mpeg2_layer1(uint8_t header_index) {
        return v2l1_brs[header_index];
    }

    inline bitrate_t get_mpeg2_layer2(uint8_t header_index) {
        return v2l2_brs[header_index];
    }

    inline bitrate_t get_mpeg2_layer3(uint8_t header_index) {
        return get_mpeg2_layer2(header_index);
    }
}

namespace mp3plot {

namespace mp3 {

// All methods in standard_bitrates are static

#if 0
bool standard_bitrates::init_done = false;
standard_bitrates::bitrate_vector standard_bitrates::bitrates;
#endif

bitrate_t bitrate_translator::get(uint8_t brindex, mpeg_version_t ver, mpeg_layer_t lyr) {
    if (ver == MPEG1) {
        if (lyr == LAYER_III)
            return get_mpeg1_layer3(brindex);
        if (lyr == LAYER_II)
            return get_mpeg1_layer2(brindex);
        if (lyr == LAYER_I)
            return get_mpeg1_layer1(brindex);
    }
    else if (ver == MPEG2 || ver == MPEG2_5) {
        if (lyr == LAYER_III)
            return get_mpeg2_layer3(brindex);
        if (lyr == LAYER_II)
            return get_mpeg2_layer2(brindex);
        if (lyr == LAYER_I)
            return get_mpeg2_layer1(brindex);
    }
    // XXX: What to do?
    // Since broken files might produce bad versions, return the MP3 value, just in case
    return get_mpeg1_layer3(brindex);
}
#if 0
void standard_bitrates::init() {
    if (init_done) return;

    // Invariant
    assert( bad_bitrate() < 0 );

    // The indices match the 4-bit header indication of bitrate
    // FIXME: This table is only valid for Layer III, see references
    static const bitrate_t BITRATES[] = {
        0,      // Free-form bitrate. Free form streams are CBR and no indication
                // of the bitrate is made explicitly, so it must be derived from the
                // frame size.
                // Although they're CBR, it is legal (or at least not-uncommon) for
                // mp3 files to be concatenated so a change in size is not impossible
        32,
        40,
        48,
        56,
        64,
        80,
        96,
        112,
        128,
        160,
        192,
        224,
        256,
        320,
        bad_bitrate(), // Bad bitrate, the frame header cannot contain a bad bitrate
                       // identifier, its presence is an indication of an error
    };

    bitrates = std::vector<bitrate_t>(BITRATES, BITRATES+sizeof(BITRATES)/sizeof(bitrate_t));

    init_done = true;
}

const std::vector<bitrate_t> standard_bitrates::as_vector() {
    init();
    return bitrates;
}

bitrate_t standard_bitrates::bad_bitrate() {
    assert( ((bitrate_t)0xFF) != ((bitrate_t)-1) );
    return -1;
}

bitrate_t standard_bitrates::get(uint8_t bitrate_header_id) {
    assert( as_vector()[bitrate_header_id] == get_mpeg1_layer3(bitrate_header_id) );
    return as_vector()[ bitrate_header_id ];
}

std::ostream & operator<<(std::ostream & os, const standard_bitrates::bitrate_vector & v) {
    int n=0;
    for (std::vector<bitrate_t>::const_iterator it = v.begin();
         it != v.end();
         ++it)
    {
        os << "[" << std::bitset<4>(n++) << "] = " << *it;
        if (*it == standard_bitrates::bad_bitrate()) {
            os << " (invalid)";
        }
        else if (*it == 0) {
            os << " (free-form)";
        }
        os << "\n";
    }

    return os;
}
#endif
} // namespace mp3

} // namespace mp3plot

// vim:set ts=4 et ai:

