/*
 * mp3plot - Bitrate analysis tool
 *
 * Copyright (C) 2007, 2009 Toni Corvera
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

// $Id: frame.cc 1347 2009-06-22 15:53:15Z toni $

#include <iostream>
#include <cassert>
#include <cstring> // memcmp

#include "../mp3_config.h" // btoh32
#include "frame.h"

// References:
// [[HA1]] Frame Size, Calc bitrate, length, vbr/cbr etc..
//         <http://www.hydrogenaudio.org/forums/index.php?showtopic=52216>
// [[HA2]] Is this a MP3 Frame header
//         <http://www.hydrogenaudio.org/forums/index.php?showtopic=43172>

using std::ifstream;
using std::string;
using std::ios;
using std::cout;

namespace {
    // These are used to avoid showing a warning for each frame tainting the results
    int unsupported_frequency = 0;
    int bad_bitrate = 0;
    int freeform_bitrate = 0;
    int mpeg25 = 0;
}

namespace mp3plot {

namespace mp3 {

    using net_outlyer::btoh32;

// -- MP3 Frame --

void frame::read(ifstream & in) throw (e_error) {
    using std::cerr;
    using std::endl;

    uint8_t sync_header[4];
    in.read(CHARR(sync_header), 4);
    if (!in.good()) {
        throw e_read_failure();
    }

    // The sync part is 12 bits, starting on byte-boundary
    // ((11 bits for the non-standard MPEG 2.5) set to 1
    if (sync_header[0] != 0xff || (sync_header[1]&0xf0)!=0xf0) {
       throw e_read_failure("Invalid frame start.", EX_DATAERR);
    }

    assert( in.tellg() >= 4 );
    start = -4;
    start += in.tellg(); // compiler complains if done in one operation

    // MPEG Version
    {
        // 2nd byte: xxxY Yxxx
        uint8_t version_field = (sync_header[1] & 0x18) >> 3;
        mpeg_version = mpeg_version_t(version_field);
        // TODO: My MPEG2.5 sample doesn't seem to match this, might be a sync issue
        if (MPEG2_5 == mpeg_version && 0 == mpeg25) {
            // FIXME: Add support
            cerr << "WARNING: MPEG 2.5 extensions detected. Information *WILL* be innacurate." << endl;
            ++mpeg25;
        }

        // 2nd byte: xxxx xYYx
        uint8_t layer_field = (sync_header[1] & 0x06) >> 1;
        mpeg_layer = mpeg_layer_t(layer_field);

        // 2nd byte: xxxx xxxY
        // This works in reverse, 1 means NO CRC
        crc = 1 != (sync_header[1] & 0x01);

        // 3rd byte: YYYY xxxx
        bitrate_index = ( (sync_header[2] & 0xf0) >> 4 );
        assert( bitrate_index < 0x10 );
        bitrate_ = bitrate_translator::get(bitrate_index, mpeg_version, mpeg_layer);

        // 3rd byte: xxxx YYxx
        uint8_t freq_field = (sync_header[2] & 0x0c) >> 2;
        switch (freq_field) { // FIXME: The table is different for MPEG-2
            case 0: freq=44100; break;
            case 1: freq=48000; break;
            case 2: freq=32000; break;
            case 3: freq=0; break; // reserved
            default:
                    assert( false );
                    freq=-1;
            ;
        }

        // Adjust according to MPEG version
        if (MPEG2_5 == mpeg_version) {
            freq /= 4;
        }
        else if (MPEG2 == mpeg_version) {
            freq /= 2;
        }

        // 3rd byte: xxxx xxYx
        has_padding = 0 != (sync_header[2] & 0x02);

        // 3rd byte: xxxx xxxY
        // Private bit, can be safely discarded

        // 4th byte: YYxx xxxx
        channel_mode = static_cast<channel_mode_t>( (sync_header[3] & 0xc0)>>6 );

        if (bitrate_ == 0) {
            if (0 == freeform_bitrate) {
                cerr << "WARNING: Free-format or corrupted bitstream detected.\n"
                        "\tInformation might be innacurate." << endl;
            }
            ++freeform_bitrate;
        }
        else if (bitrate_ == bitrate_translator::bad_bitrate()) {
            if (0 == bad_bitrate) {
                cerr << "WARNING: Illegal value found in the bitrate field.\n"
                        "\tThis is usually a sign of a corrupted file.\n"
                        "\tInformation might be innacurate." << endl;
            }
            ++bad_bitrate;
        }

#if 0 // FIXME: This breaks the frame detection. Why?
        if (crc) {
            // CRC is stored as 16bits immediately following the header
            in.seekg(2, ios::cur);
        }
#endif
        // TODO: parse the other fields
    }
}

void frame::skip_data(ifstream & in) throw (e_read_failure) {
    using std::cerr;
    using std::endl;

    uint16_t f = freq;

    if (f == 0) {
        f = 44100;
        if (0 == unsupported_frequency) {
            cerr << "WARNING: Unsupported frequency found, using a sane default.\n"
                    "\tInformation might be innacurate.\n"
                    "\tThis warning won't be reported again." << endl;
        }
        ++unsupported_frequency;
    }

    // Per PRE not allowed
    assert( bitrate_translator::bad_bitrate() != bitrate_ );
    if (bitrate_translator::bad_bitrate() == bitrate_) {
        throw e_read_failure("Attempted to read a frame with bad bitrate");
    }

    // Ditto
    assert( 0 != bitrate_ );
    if (0 == bitrate_) {
        throw e_read_failure("Attempted to read a freeform bitstream frame");
    }


    const uint16_t factor = get_sample_size() / 8;
    assert( factor==48 || factor == 72 || factor == 144 );

    uint32_t framelen = int(factor * (bitrate_ * 1000) / f ) -4;
    if (has_padding) {
        // In layer I padding is 4 bytes (that's why the formula is usually given as
        // '(12 * br / freq + pad) * 4')
        if (mpeg_layer == LAYER_I) {
            framelen += 4;
        }
        else {
            ++framelen;
        }
    }

    in.seekg(framelen, ios::cur);
}

// See [[HA1]], [[HA2]]
// static
uint16_t frame::get_sample_size(mpeg_version_t v, mpeg_layer_t l) {
    // Layer I (MPEG 1 and 2) has 384 samples per frame,
    // MPEG-1 Layer II and III and MPEG-2 Layer 2 have 1152 samples per frame
    // MPEG-2 Layer II has 576 samples per frame

    if (l == LAYER_I) {
        // MPEG-1/2 Layer I
        return 384;
    }
    if (l == LAYER_III && v == MPEG2) {
        // MPEG-2 LSF Layer III
        return 576;
    }
    // MPEG-1 Layers II and III, MPEG-2 Layer II
    return 1152;
}

void frame_first::read(ifstream & in) throw (e_error) {
    frame::read(in);

    using std::cerr;

    // POST ]-> At end of frame header (4th byte, [3])

    check_vbr(in);
    if (!info.is_vbr) {
        // Note even CBR files can have some VBR due to bit reservoir
        cerr << "This appears to be a CBR file" << std::endl;
    }

    // POST ]-> At 41th frame byte ([40]) || !is_vbr

    { // TODO: Is this right?
        in.seekg(3, std::ios_base::cur);
        in.read(CHARR(&flags), 1);
    }

    info.has_filesize = (0 != (flags & FL_FILESIZE));
    info.has_numframes = (0 != (flags & FL_NUMFRAMES));
    info.has_toc = (0 != (flags & FL_TOC));
    info.has_vbr_scale = (0 != (flags & FL_VBR_SCALE));

    // POST ]-> At 45th frame byte ([44])
    {
        if (info.has_numframes) {
            in.read(CHARR(&info.num_frames), 4);
            // XXX: is this ok in all arches?
            info.num_frames = btoh32(info.num_frames);
        }
        else {
            in.seekg(4, ios::cur);
        }

        if (info.has_filesize) {
            in.read(CHARR(&info.file_size), 4);
            info.file_size = btoh32(info.file_size);
        }
        else {
            in.seekg(4, ios::cur);
        }

        // FIXME: Do something with the TOC ?
        in.seekg(99, ios::cur);
        // FIXME: Do something with VBR Scale ?
        in.seekg(4, ios::cur);
    }
}

void frame_first::check_vbr(ifstream & in) throw (e_read_failure) {
    // PRE ]-> At 4th byte [3]

    const int BLOCK=36;
    uint8_t buffer[BLOCK];
    in.read(CHARR(buffer), BLOCK);

    int offset=-1;
    while (!info.is_vbr && ++offset < (BLOCK-3)) {
        uint8_t * ptr = & (buffer[offset]);

        if (0 == memcmp(ptr, "Xing", 4)) {
            info.is_vbr = true;
            xing_offset = offset + 4; // We started after the first four bytes
        }
    }
    // FIXME: do validity check of offset
    // // <http://www.multiweb.cz/twoinches/MP3inside.htm>
    // 36		"Xing" for MPEG1 and CHANNEL != mono (most used)
    // 21		"Xing" for MPEG1 and CHANNEL == mono
    // 21		"Xing" for MPEG2 and CHANNEL != mono
    // 13		"Xing" for MPEG2 and CHANNEL == mono

    if (!info.is_vbr) { // Return to the CBR frame start
        in.seekg( -40, std::ios_base::cur );
    }
}

std::ostream & operator<<(std::ostream & os, const frame & f) {
    string ver, layer, chanmode;

    switch (f.mpeg_version) {
        case MPEG1: ver="MPEG v1"; break;
        case MPEG2: ver="MPEG v2"; break;
        case MPEG2_5: ver="MPEG v2.5"; break;
        default: ver="undefined";
    }
    switch (f.mpeg_layer) {
        case LAYER_III: layer="Layer III"; break;
        case LAYER_II: layer="Layer II"; break;
        case LAYER_I: layer="Layer I"; break;
        default: layer="undefined";
    }
    switch (f.channel_mode) {
        case STEREO: chanmode="Stereo"; break;
        case JOINT_STEREO: chanmode="Joint Stereo"; break;
        case DUAL_MONO: chanmode="Dual Mono"; break;
        case MONO: chanmode="Mono"; break;
        default: chanmode="undefined"; break;
    }

    os << "MPEG Version: " << ver << "\n";
    os << "MPEG Layer: " << layer << "\n";
    os << "Has CRC: " << (f.crc?"Yes":"No") << "\n";
    // TODO: Do something on VBR files, this is always 128 and has no meaning AFAIK
    os << "Nominal bitrate: " << f.bitrate_ << " kbps\n";
    os << "Sampling rate: " << f.freq << " Hz\n";
    os << "Channel mode: " << chanmode << "\n";

    return os;
}

} // namespace mp3

} // namespace mp3plot

// vim:set ts=4 et ai:

