/*
 * mp3plot - Bitrate analysis tool
 *
 * Copyright (C) 2007, 2009 Toni Corvera
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

// $Id: frame.h 1347 2009-06-22 15:53:15Z toni $

#ifndef MP3PLOT_MP3_FRAME_H
#define MP3PLOT_MP3_FRAME_H

#include <string>
#include <fstream>

#include "element.h"
#include "bitrate.h"

namespace mp3plot {

namespace mp3 {

    using std::ostream;

    // Forward declaration
    struct meta_info;

    // Forward declarations (MSVC requires them)
    class frame;
    ostream & operator<<(ostream &, const frame &);

    /*!
     * \brief MP3 audio frame and associated header.
     */
    class frame : public mp3::element {
    public:
        // output operator
        friend std::ostream & mp3plot::mp3::operator<<(std::ostream &, const frame &);

        explicit frame()
            : mpeg_version(MPEG_VERSION_RESERVED),
              mpeg_layer(LAYER_RESERVED),
              crc(false),
              bitrate_(bitrate_translator::bad_bitrate()),
              freq(-1),
              has_padding(false),
              channel_mode(CHANNEL_MODE_UNDEFINED),
              bitrate_index(0x0F), // BAD
              start(-1)
        {}
        virtual ~frame() {}

        /*!
         * \brief Reads the frame meta data and puts the get pointer in the audio data.
         * \see skip_data()
         * \param input The input file from which to read
         * \throw e_error If not at a frame start
         * \throw e_error If a read error occurs
         * \post Read pointer is at the start of the audio data
         */
        virtual void read(std::ifstream & input) throw (e_error);

        /*!
         * \brief Skips the audio data.
         * \note The amount skipped is based on bitrate, frequency and
         *       known standard values.
         * \param input The input file from which to read
         * \throw e_read_failure If a read error occurs
         * \pre Input stream is pointed at the start of the frame's audio data
         * \pre The frame has a non-bad bitrate (will throw otherwise)
         * \pre The frame has a non-free bitrate (will throw). Note, this precondition
         *      will be removed in the future.
         * \todo TODO: Allow skipping of free-form
         */
        void skip_data(std::ifstream & input) throw (e_read_failure);

        //! Returns the offset of the frame start in the file.
        const std::streampos & offset() const { return start; }

        //! Bitrate getter
        bitrate_t bitrate() const {  return bitrate_; }

        // Since 0.5.2, used in the transitional support for other MPEG versions/layer
        // TODO: Cleanup the interface
        mpeg_version_t get_mpeg_version() const  { return mpeg_version; }
        mpeg_layer_t get_mpeg_layer() const { return mpeg_layer; }
        uint8_t get_bitrate_index() const { return bitrate_index; }
        uint16_t get_sampling_rate() const { return freq; }

        static uint16_t get_sample_size(mpeg_version_t, mpeg_layer_t);
        uint16_t get_sample_size() const {
            return get_sample_size(mpeg_version, mpeg_layer);
        }

    protected:
        /*!
         * \brief Index of the frame in the file.
         * \deprecated Only used for debugging purposes
         * \todo TODO: To be removed
         */
        numframes_t frame_idx;

        /// Header meta information

        //! Version of MPEG (1, 2 or 2.5) of the frame
        mpeg_version_t mpeg_version;
        //! MPEG Layer (I, II or III) of the frame
        mpeg_layer_t   mpeg_layer;
        //! true if the frame contains a CRC
        bool           crc;
        //! Bitrate, in kbps, of the frame
        bitrate_t      bitrate_;
        //! Frequency, in Hz, of the frame
        uint16_t       freq;
        //! true if the frame has padding
        bool           has_padding;
        //! Channel mode of the frame
        channel_mode_t channel_mode;

        //! Bitrate index as found on the headers
        uint8_t bitrate_index;

    private:
        /*!
         * \brief Offset of the frame start
         * \see offset()
         */
        std::streampos start;
    };

    /*!
     * \brief Represents the first frame.
     * The first frame is a special case of frame <em>in VBR files</em>.
     * \see frame
     */
    class frame_first : public frame {
    public:
        //! Default constructor
        explicit frame_first(meta_info & meta_info, bool is_tainted_)
            : frame(),
              xing_offset(0),
              flags(0),
              info(meta_info),
              tainted(is_tainted_)
        {}

        /*!
         * \brief Read the frame information.
         * \copydoc frame::read()
         */
        virtual void read(std::ifstream & input) throw (e_error);

        //! True if this is a tainted first frame
        bool is_tainted() const { return tainted; }

    private:
        /*!
         * \brief Checks if the file is VBR.
         * \throw e_read_failure If a read error occurs
         */
        void check_vbr(std::ifstream & in) throw (e_read_failure);

        //! Byte offset of the xing string (see MP3 specs)
        std::streampos xing_offset;
        //! Flags of the frame (unused)
        uint8_t flags;
        //! File metainformation stored in the first frame if a VBR file
        meta_info & info;
        /*!
         * \brief When true, information might be less accurate
         * When true, the frame is the first found but might not be the actual first
         * (e.g. on broken files)
         */
        bool tainted;
    };

    /*!
     * \brief Meta information of the file
     * \note This is the information located in the first frame of VBR files
     */
    struct meta_info {
        bool is_vbr;
        uint32_t num_frames;
        // This is the filesize indicated in the header, if any, or
        // 0 otherwise, apparently it might differ somewhat from its real size
        uint32_t file_size;

        bool has_filesize, has_numframes, has_toc, has_vbr_scale;

        meta_info()
            : is_vbr(false),
              num_frames(0),
              file_size(0),
              has_filesize(false),
              has_numframes(false),
              has_toc(false),
              has_vbr_scale(false)
        {}
    };

    // Forward declaration for documentation purposes
    /*!
     * \brief Prints the frame meta information
     */
    std::ostream & operator<<(std::ostream &, const frame &);

} // namespace mp3

} // namespace mp3plot

#endif // MP3PLOT_MP3_FRAME_H
// vim:set ts=4 et ai:

