/*
 * mp3plot - Bitrate analysis tool
 *
 * Copyright (C) 2007, 2009 Toni Corvera
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

// $Id: mp3_file.cc 1347 2009-06-22 15:53:15Z toni $

#include "mp3_file.h"

#include <iostream>
#include <cassert>
#include <cstring> // memcmp

#include "utils/str_utils.h"
#include "spinner.h"

/*!
 * \file
 * \remarks File structure based on <http://www.multiweb.cz/twoinches/MP3inside.htm>
 */

namespace mp3plot {

using std::ifstream;
using std::cerr;
using std::clog;
using std::cout;
using std::endl;
using std::string;
using std::ios;

// -- MP3 File --

mp3file::mp3file(const string & file_path)
    : file_type(UNKNOWN),
      meta_info_(),
      file(file_path.c_str(), ios::binary|ios::in),
      junk_length(0)
{}

void mp3file::check_file_type() throw (e_error) {
    file.seekg(0, ios::beg);

    uint8_t filestart[3];
    file.read(CHARR(filestart), 3);

    if (0 == ::memcmp(filestart, "ID3", 3)) {
        file_type = FT_ID3V2;
    }
    else if (filestart[0] == 0xff && 0xf0 == (filestart[1] & 0xf0)) {
        file_type = FT_MP3;
    }
    else {
        try {
            seek_to_next_frame();
        }
        catch (const e_read_failure & e) {
            cerr << "End of file reached without finding a frame\n\t" << e.what() << endl;
            throw e_bad_filetype();
        }
        junk_length = file.tellg();
        // If there's too much junk this might not be an mp3 file. Since
        // no file can safely be discarded simply warn the user
        if (junk_length > MAX_JUNK_THRESHOLD) {
            cerr << "WARNING: A lot of junk found at the beginning og the file (" <<
                    junk_length << ")\n"
                    "\tThis might not be an MP3 file, continuing anyway but information\n"
                    "\tmight be innacurate." << endl;
        }
        else {
            cerr << "WARNING: Junk at the beginning of the file (up to byte #" <<
                    junk_length << ")\n"
                    "\tFile information might be inaccurate.\n" << endl;
        }
    }
}

void mp3file::read_info() throw (e_error) {
    check_file_type();

    file.seekg(0, ios::beg);
    if (FT_ID3V2 == file_type) {
        id3v2.read(file);

        // Jump to tag end:
        file.seekg(id3v2.tag_size, ios::cur); // ID3v2 total size is header(=10) + tag_size
    }

    // Find first frame
    seek_to_next_frame(); // throws e_read_failure if none found
    // POST ]-> at first frame

    const bool tainted = 0 != junk_length;
    std::auto_ptr<mp3::frame_first> fframe (new mp3::frame_first(meta_info_, tainted));
    fframe->read(file);
    numframes_t frame_idx = 1;
    frames.push_back( *fframe );

    net_outlyer::spinner spinner;
    try {
        spinner.start();

        const bitrate_t BAD = mp3::bitrate_translator::bad_bitrate();

        while (true) {
            seek_to_next_frame();
            mp3::frame f;
            frame_idx++;
            f.read(file);
            if (f.bitrate() == BAD) {
                // Advance the file slightly, will try to sync on the next
                // iteration
                file.seekg(1, ios::cur);
            }
            else if (f.bitrate() == 0) {
                // Free-form bitrate.
                // There's no explicit field in the file indicating the bitrate,
                // it must be guessed from the frame size (hence, simply sync)

                // Skip the absolutely minimum size possible
                // 32 = 144 * 8 * 1000 / 32000 - 4 (see frame.skip_data())
                file.seekg(32, ios::cur);
            }
            else {
                f.skip_data(file);
            }
            frames.push_back(f);

            spinner++;
        }
    }
    catch (UNUSED_ const e_read_failure & e) {
        spinner.stop();
        // This isn't really an error
        clog << "No more frames found after byte #" << frames.back().offset() << "\n";
    }
    spinner.stop();
    clog << frame_idx << " frames read" << endl;
}

std::ostream & operator<<(std::ostream & os, const mp3file & file) {
    if (FT_ID3V2 == file.file_type) {
        os << file.id3v2;
    }
    else {
        os << "This file doesn't contain an ID3v2 header\n";
    }
    os << "* Information gathered from the first frame:\n" << file.frames.front();

    os << "* VBR Information: "; // no \n: CBR => inline, VBR => newline
    if (!file.meta_info_.is_vbr) {
        os << "NOT a VBR file\n";
    }
    else { // VBR
        os << "\nVBR file\n";
        if (file.meta_info_.has_numframes) {
            os << "\tNumber of frames included (" << file.meta_info_.num_frames << ")\n";
        }
        if (file.meta_info_.has_filesize)
            os << "\tFile size included (" << file.meta_info_.file_size << ")\n";
        if (file.meta_info_.has_toc)
            os << "\tTOC included\n";
        if (file.meta_info_.has_vbr_scale)
            os << "\tVBR scale included\n";
    }
    return os << std::flush;
}

void mp3file::fill_bitrate_distribution() const {
    if (br_dist.is_filled) return;

    br_dist.fill_from(frames, meta_info_.is_vbr);
    assert( br_dist.is_filled );
}

void mp3file::plot(const bitrate_plotter & brp, const plot_params& params) throw (e_error){
    fill_bitrate_distribution();

    brp.plot(*this, params); // might throw e_error
}

void mp3file::seek_to_next_frame() throw (e_read_failure) {
    uint8_t last=0, byte;

    // Frames start with { 0xFF , 0xFX }

    while (file.good()) {
        file.read(CHARR(&byte), 1);

        if ((last==0xff) && 0xf0 == (byte&0xf0)) {
            file.seekg( -2 , ios::cur);
            return;
        }

        last = byte;
    }

    throw e_read_failure("Couldn't find next frame.", EX_DATAERR);
}

} // namespace mp3plot

// vim:set ts=4 et ai:
