/*
 * mp3plot - Bitrate analysis tool
 *
 * Copyright (C) 2007, 2009 Toni Corvera
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

// $Id: bitrate_plotter.h 1344 2009-06-22 01:11:09Z toni $

#ifndef MP3PLOT_BITRATE_PLOTTER_H
#define MP3PLOT_BITRATE_PLOTTER_H

#include "../mp3_config.h"

// TODO / [C++]: auto_ptr isn't compatible with STL containers,
//                what other smart pointer might be preferable?
#include <boost/shared_ptr.hpp>

#include <string>
#include <vector>

#include "../common.h"
#include "../errors.h"
#include "../bitrate_dist.h"
#include "plotter_features.h"

namespace mp3plot {

    // Forward declaration
    class mp3file;
    struct plot_params;

    using boost::shared_ptr;

    /*!
     * \brief (optional) parameters for plotting
     * \since 0.5.0
     */
    struct plot_params {
        plot_params(const std::string& filename, const bitrate_distribution & brd)
            : output_filename(filename), w(0), h(0),
              old_colourscheme(false),
              br_dist(brd)
        {}

        const std::string output_filename;
        const int w, h;
        bool old_colourscheme; //!< \since 0.5.1
        const bitrate_distribution & br_dist; //!< \since 0.5.2
    };

    /*!
     * \brief Creates a bitrate plot based on a bitrate distribution
     * \see bitrate_distribution
     */
    class bitrate_plotter {
    private:
        // XXX? Can't use unwrapped as it is abstract
        typedef shared_ptr<bitrate_plotter> wrapper_type;
        typedef std::vector<wrapper_type> plotters_type;
        static plotters_type plotters;

    protected:
        bitrate_plotter() {}
    public:
        virtual ~bitrate_plotter() {};

        // FIXME: Error code
        struct e_unregistered_plotter_type : e_error {
            e_unregistered_plotter_type()
                : e_error("No plotter of such type", EX_SOFTWARE)
            {}
        };

        struct e_no_plotter_matches_name : e_error {
            e_no_plotter_matches_name()
                : e_error("No plotter with such name available", EX_SOFTWARE)
            {}
        };

        friend struct matches_features;
        friend struct includes_features;
        friend struct matches_name;
        template<typename Predicate_, typename T_exception_>
            friend bitrate_plotter & choose_impl(Predicate_ p)
                throw (T_exception_);

        bool matches(const wrapper_type & A) const { return true; }

        /*!
         * \brief Creates the plot
         * \param file The MP3 file, with data already red
         * \throw e_error if the plot couldn't be finished
         */
        virtual void plot(const mp3file & file, const plot_params& params) const
            throw (e_error) = 0;

        /*!
         * \brief Describes the features of this plotter
         */
        virtual const plotter_features& features() const = 0;

        virtual const plotter_description& description() const = 0;

        /*!
         * \brief Chooses a plotter with the requested features upon the
         * registered ones.
         * \throws e_unregistered_plotter_type If no plotter has the requested
         * features.
         * \see choose_fuzzy(const plotter_feature &)
         */
        static bitrate_plotter & choose(const plotter_features &)
                                        throw (e_unregistered_plotter_type);

        static bitrate_plotter & choose(const std::string & name)
                                        throw (e_no_plotter_matches_name);

        static std::vector<std::string> list();

        /*!
         * \brief Chooses a plotter with the requested features (of more)
         * upon the registered ones.
         * \throws e_unregistered_plotter_type If no plotter has at least the
         * requested features.
         * \see choose(const plotter_features &)
         */
        static bitrate_plotter & choose_fuzzy(const plotter_features &)
                                        throw (e_unregistered_plotter_type);

    private:
        /*!
         * \brief Requirement for registrations.
         * \see Evolving a language in and for the real world: C++ 1991-2006, page 59
         *      http://www.research.att.com/~bs/papers.html
         */
        template<typename T_>
        struct constraint_is_plotter {
            static void constraint(T_ a) {
               UNUSED_ bitrate_plotter & X = a;
            }

            constraint_is_plotter() {
               UNUSED_ void (*p) (T_) = constraint;
            }
        };

    public:
        template<typename T_plotter> static void register_plotter() {
            constraint_is_plotter<T_plotter>();

            wrapper_type p(new T_plotter());
            plotters.push_back( p );
        }
    };

} // namespace mp3plot

#endif // MP3PLOT_BITRATE_PLOTTER_H
// vim:set ts=4 et ai:

