/*
 * Copyright (C) 2007 Toni Corvera
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

// $Id: colour.cc 1113 2009-04-09 01:33:39Z  $

#include "colour.h"

namespace {
    /*!
     * \brief Abstract implementation of the output operator
     * \param T_scalar    Type of the components
     * \param T_cast_type Components will be casted to this type, if no cast
     *                    is required, use template<typename,int>print_col()
     * \param C_max       Full component value
     */
    template<typename T_scalar, typename T_cast_type, int C_max>
    inline std::ostream & print_col(std::ostream & os,
                                    const T_scalar & r,
                                    const T_scalar & g,
                                    const T_scalar & b)
    {
        return os << "Col[0.." << C_max << "] ("
                         << static_cast<T_cast_type>(r) << ", "
                         << static_cast<T_cast_type>(g) << ", "
                         << static_cast<T_cast_type>(b) << ")";
    }

    /*!
     * \brief Short-hand of print_col, when no typecast is required.
     */
    template<typename T_scalar, int C_max>
    inline std::ostream & print_col(std::ostream & os,
                                    const T_scalar & r,
                                    const T_scalar & g,
                                    const T_scalar & b)
    {
        return print_col<T_scalar, T_scalar, C_max>(os, r,g,b);
    }
}

namespace net_outlyer {

/////////////////////////////////////////////
// Floating point implementations [0..1]
/////////////////////////////////////////////

std::ostream & operator<<(std::ostream & os, const colour_type<long double,1> & c) {
    return print_col<long double,1>(os, c.r, c.g, c.b);
}

std::ostream & operator<<(std::ostream & os, const colour_type<double,1> & c) {
    return print_col<double,1>(os, c.r, c.g, c.b);
}

std::ostream & operator<<(std::ostream & os, const colour_type<float,1> & c) {
    return print_col<float,1>(os, c.r, c.g, c.b);
}

/////////////////////////////////////////////
// Integer implementations [0..255]
/////////////////////////////////////////////

std::ostream & operator<<(std::ostream & os, const colour_type<uint8_t,255> & c) {
    return print_col<uint8_t, int, 255>(os, c.r, c.g, c.b);
}

std::ostream & operator<<(std::ostream & os, const colour_type<signed short,255> & c) {
    return print_col<signed short, 255>(os, c.r, c.g, c.b);
}

std::ostream & operator<<(std::ostream & os, const colour_type<unsigned short,255> & c) {
    return print_col<unsigned short, 255>(os, c.r, c.g, c.b);
}

std::ostream & operator<<(std::ostream & os, const colour_type<signed int,255> & c) {
    return print_col<signed int, 255>(os, c.r, c.g, c.b);
}
std::ostream & operator<<(std::ostream & os, const colour_type<unsigned int,255> & c) {
    return print_col<unsigned int, 255>(os, c.r, c.g, c.b);
}

}

// vim:set ts=4 et ai:
