/*
 * Copyright (C) 2007, 2009 Toni Corvera
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

// $Id: colour.h 1211 2009-04-24 19:06:20Z  $

#ifndef ONET_COLOUR_H
#define ONET_COLOUR_H

#include <iostream>

#ifdef _MSC_VER
#	include "../msvc/inttypes.h"
#else
#	include <inttypes.h>
#endif

#include "../utils/constraints.hxx"

namespace net_outlyer {

    /*!
     * \brief Representation of an RGB colour.
     * There are different implementations of RGB colours, the two most
     * common being an 8bit integer per component one and a floating
     * point in the range 0..1 per component.
     * This class tries to abstract their differences: the aforementioned
     * types match
     * \code colour_type<uint8_t, 255> \endcode (but see remarks)
     * and \code colour_type<double, 1> \endcode
     * \remarks The range validity is deliberately left out of the
     *          constructors, users are allowed to go beyond the bounds
     * \remarks As this class contains (optional) checks for ranges, the
     *          integer 0..255 representation may throw warnings when
     *          implemented over an 8bit type (can't be out of range =>
     *          check is always false).
     * \remarks Conversion between representations is provided in the convert()
     *          method, see its documentation for details on performance though.
     * \param T_component Type of the colour components
     * \param C_max Value of the full component (maximum colour intensity)
     */
    template<typename T_component, int C_max>
    struct colour_type {
    public:
        T_component r, //!< Red component
                    g, //!< Green component
                    b; //!< Blue component

        typedef T_component value_type;

        /*!
         * \brief Standard constructor.
         * \param[in] r_ Red component
         * \param[in] g_ Green component
         * \param[in] b_ Blue component
         * \sa check_bounds()
         */
        colour_type(T_component r_, T_component g_, T_component b_)
            : r(r_), g(g_), b(b_)
        {
            constraints::static_casts_to<T_component, double>();
        }

        /*!
         * \brief Gray-scale constructor.
         * Colours in the gray-scale are colours with all three components
         * equal. Such colours represent levels of gray.
         * \param[in] gl Gray level
         * \sa check_bounds()
         */
        explicit colour_type(T_component gl)
            : r(gl), g(gl), b(gl)
        {
            constraints::static_casts_to<T_component, double>();
        }

        /*!
         * \brief Default constructor.
         * Initialises the colour components to 0.
         */
        colour_type()
            : r(0), g(0), b(0)
        {
            constraints::static_casts_to<T_component, double>();
        }

        /*!
         * \brief Copy constructor.
         * \remarks Different colour implementations can't be directy copied,
         *          but they can through conversion, e.g.:
         * \code
         * colour_type<double, 1>   dcol;
         * ...
         * colour_type<int, 255> icol( dcol.convert<int, 255>() );
         * \endcode
         */
        colour_type(const colour_type<T_component, C_max> & o)
            : r(o.r), g(o.g), b(o.b)
        {}

        /*!
         * \brief Assignment operator.
         * \pre  <code> THIS = *this and O = o </code>
         * \post <code> *this == O </code>
         */
        colour_type<T_component, C_max> & operator=(const colour_type<T_component, C_max> & o) {
            if (this == &o) { return *this; }

            r = o.r, g = o.g, b = o.b;
            return * this;
        }

        /*!
         * \brief Equivalence operator.
         * \remarks Note that the strength of this comparison depends on the
         *          strength of \p T_component 's comparison operator (e.g.
         *          two real-world equivalent doubles might not be equivalent
         *          to the machine).
         */
        bool operator==(const colour_type<T_component, C_max> & c) const {
            if (this == &c) { return true; }

            return c.r = r && c.g == c.g && c.b == c.b;
        }

        /*!
         * \brief Inequivalence operator.
         * \pre  <code> THIS = *this AND C = c </code>
         * \post <code> THIS != C  ==  !(THIS == C) </code>
         */
        bool operator!=(const colour_type<T_component, C_max> & c) const {
            return !(*this == c);
        }

        /*!
         * \brief Conversion.
         * This method converts between implementations.
         * \remarks This method uses floating point intermediate values to do
         *          the conversion, so although <code>O(1)</code>, it isn't
         *          completely cheap. This is of special interest when
         *          the converted colour's component must be accesed, in which
         *          case it's better to hold the conversion in a temporal
         *          variable.
         * \param T_C_component Type of the target colour components
         * \param C_C_max Value of the target's full component (maximum colour
         *                intensity)
         * \post Let \p T1 and \p T2 be scalar types with perfect == and !=
         *       operators, \p X, \p Y integer values, and either
         *       \p T1!=T2 or \p X!=Y (or both):
         *
         *      \code
         * colour_type<T1, X> c1, d;
         * colour_type<T2, Y> c2;
         * ...
         * c2 = c1.convert<T2, Y>();
         * d = c2.convert<T1, X>();
         * // d and c1 must be equal
         * assert( d == c1 );
         *      \endcode
         * \todo FIXME: There's a problem with long double components
         */
        template<typename T_C_component, int C_C_max>
        colour_type<T_C_component, C_C_max> convert() const {
            // Need to be sure the intermediates are FP:
            const double
                r_ = static_cast<double>(r) / C_max * C_C_max,
                g_ = static_cast<double>(g) / C_max * C_C_max,
                b_ = static_cast<double>(b) / C_max * C_C_max;

            return colour_type<T_C_component, C_C_max>(
                               static_cast<T_C_component>(r_),
                               static_cast<T_C_component>(g_),
                               static_cast<T_C_component>(b_)
                   );
        }
    }; // struct colour_type

    // Forward declarations of common output operators

    // FP implementations
    std::ostream & operator<<(std::ostream &, const colour_type<long double,1> &);
    std::ostream & operator<<(std::ostream &, const colour_type<double,1> &);
    std::ostream & operator<<(std::ostream &, const colour_type<float,1> &);

    // Integer implementations
    std::ostream & operator<<(std::ostream &, const colour_type<uint8_t,255> &);
    std::ostream & operator<<(std::ostream &, const colour_type<signed short,255> &);
    std::ostream & operator<<(std::ostream &, const colour_type<unsigned short,255> &);
    std::ostream & operator<<(std::ostream &, const colour_type<signed int,255> &);
    std::ostream & operator<<(std::ostream &, const colour_type<unsigned int,255> &);

    typedef colour_type<double,1> rgb_type;

    /*!
     * \brief HSV colour representation
     * \note Code based on <http://www.deadbeef.com/index.php/converting_rgb_to_hsv_in_c>
     * and <http://www.codeproject.com/KB/miscctrl/CPicker.aspx>
     * Hue is 0..360
     * Saturation is 0..100
     * Value is 0..100
     * \since 0.5.1
     */
    struct hsv_colour {
    public:
        double h, s, v;

        // TODO: Implement conversion to RGB
        // It is pretty tricky, so won't implement unless absolutely needed
        hsv_colour(const colour_type<double,1>& c) {

            const double maxC = std::max(std::max(c.r, c.g), c.b);
            const double minC = std::min(std::min(c.r, c.g), c.b);
            const double delta = maxC - minC;

            if (0 == delta) {
                h = 0;
                s = 0;
                v = maxC;
            }
            else {
                v = maxC * 100.0;
                s = (delta / maxC) * 100.0;

                const double dR = 60*(maxC - c.r)/delta + 180;
                const double dG = 60*(maxC - c.g)/delta + 180;
                const double dB = 60*(maxC - c.b)/delta + 180;

                if (maxC == c.r) {
                    h = dB - dG;
                }
                else if (maxC == c.g) {
                    h = 120 + dR - dB;
                }
                else {
                    h = 240 + dG - dR;
                }

                if (0 > h) {
                    h += 360.0;
                }
                else if (h >= 360) {
                    h -= 360.0;
                }
            }
        }
    }; // hsv_colour

} // namespace net_outlyer

#endif // ONET_COLOUR_H
// vim:set ts=4 et ai:

