/*
 * mp3plot - Bitrate analysis tool
 *
 * Copyright (C) 2007, 2009 Toni Corvera
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

// $Id: console_plotter.cc 1347 2009-06-22 15:53:15Z toni $

// References:
// [[IMP3C]] Inside the MP3 Codec - Page 13
//           <http://www.mp3-converter.com/mp3codec/frames.htm>

#include <iostream>
#include <string>

#include "../mp3_config.h" // PLOT_SYMBOL
#include "../utils/str_utils.h"
#include "console_plotter.h"

using std::cout;
using std::endl;
using std::string;

namespace mp3plot {

namespace str = net_outlyer::str;

const plotter_features & console_plotter::features() const {
    static const plotter_features f ( plotter_features::textual() |
                                        plotter_features::printed() );
    return f;
}

const plotter_description & console_plotter::description() const {
    static const plotter_description D("console", features());
    return D;
}

void console_plotter::plot(const mp3file & file, const plot_params& params) const
    throw (e_error)
{
    using str::to_string;

    const bitrate_distribution & dist = params.br_dist;

    // Max length of bitrate tags
    static const size_t LBR = string("invalid").length();
    // Max length of number of frames, + padding
    const size_t LOC = str::to_string(dist.max).length() + 1;

    // Console width
    // FIXME: detect
    const uint16_t CONWIDTH = 80;
    // 4 = "[]: " + 1 => space between number of frames and dots
    // using float because a cast would be made anyway
    const float AVAIL = static_cast<float>(CONWIDTH - (LBR + LOC + 5));

    uint32_t sum=0, fs=0;
    float sumbrs=.0;

    // TODO: Adapt to changing mpeg versions/layers
    const mp3::frame & f = * file.get_frames().begin();
    const mpeg_version_t ver = f.get_mpeg_version();
    const mpeg_layer_t layer = f.get_mpeg_layer();

    for (bitrate_distribution::const_iterator b_it = dist.begin();
         b_it != dist.end();
         ++b_it)
    {
        const uint8_t bridx = b_it.bitrate_index();
        const bitrate_t br = mp3::bitrate_translator::get(bridx, ver, layer);
        const numframes_t nf = b_it.numframes();

        const string & s_br = (br != 0)
                                ? str::pad( to_string(br) + "kbps", LBR, " ", str::PAD_LEFT )
                                : str::pad( "free", LBR, " ", str::PAD_LEFT );
        const string & s_nf = str::pad( str::to_string(nf), LOC );

        cout << "[" << s_br << "]: " << s_nf;

        // the scale is relative to the max
        const int SCALED = int(nf * AVAIL / dist.max);
        cout << str::pad("", SCALED, PLOT_SYMBOL);

        cout << "\n";

        sum += nf; // number of frames of this bitrate
        sumbrs += (nf * br);
        // That's the number of bits per seconds at this bitrate
        const uint32_t secbits = nf * br * 1000;
        // Each frame is 0.026 seconds (26ms) long, ~38.46fps, see [[IMP3C]]
        fs+=int( ((double)secbits) * 0.026);
    }
    { // Finally print the invalid frames
        cout << "[invalid]: " << dist.bad_frames();
        /*
        const int SCALED = int(br_dist.bad_frames() * AVAIL / br_dist.max);
        cout << str::pad("", SCALED, PLOT_SYMBOL);
        */
        cout << "\n";
    }

    string abr = str::to_string(sumbrs/sum);
    const string::size_type dotidx = abr.find_first_of('.');
    if (string::npos != dotidx) {
        if ((dotidx + 3) < abr.length()) {
            abr.erase(dotidx+3);
        }
    }

    double length_secs = sum * f.get_sample_size() / f.get_sampling_rate();

    cout << "Average bitrate is " << abr << " kbps\n";
    cout << "\twith a total of " << sum << " frames" << endl;
    cout << "\twhich yields a theoretical (raw audio data) size of "
         << (fs / 8) << " bytes\n"
         << "\tand a length of " << length_secs << "s" << endl;
}

} // namespace mp3plot

// vim:set ts=4 et ai:

