/*
 * mp3plot - Bitrate analysis tool
 *
 * Copyright (C) 2007, 2009 Toni Corvera
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

// $Id: gd_image.cc 1344 2009-06-22 01:11:09Z toni $

#include "gd_image.h"

#if defined(HAVE_GD) && !defined(DISABLE_GRAPHICS) && !defined(DISABLE_GD)

#include <algorithm>
#include <gdfontl.h>
#include <cassert>
// FIXME: Discard gdfont* variants and use gdImageStringFT (see index.html in
//        gd packaging) OR do it optionally, in case GD isn't compiled with FT2
//        support

namespace mp3plot {

gd_image::gd_image(int w, int h, const gd_image::rgb24b_colour & c)
    throw (e_error)
: im_(::gdImageCreateTrueColor(w, h)),
  w_(w),
  h_(h),
  bgcol_(c)
{ // gdImageCreate...() returns NULL on failure
    if (nullptr == im_) {
        throw e_error();
    }
    flood( allocate_colour(bgcol_.r, bgcol_.g, bgcol_.b) );
}

// Copy constructor
gd_image::gd_image(const gd_image & gdi)
    throw (e_error)
: im_(::gdImageCreateTrueColor(gdi.w_, gdi.h_)),
  w_(gdi.w_),
  h_(gdi.h_),
  bgcol_(gdi.bgcol_)
{
    if (nullptr == im_) {
        throw e_error();
    }
    ::gdImageCopy(im_, gdi.im_, 0, 0, 0, 0, w_, h_);
}

gd_image & gd_image::operator=(const gd_image & gdi)
    throw (e_error)
{
    gd_image temp(gdi);
    // POST: temp is a full copy of gdi
    w_ = temp.w_;

    // Deallocate the current image if any
    if (0 != im_) {
        ::gdImageDestroy(im_);
    }

    h_ = temp.h_;
    bgcol_ = temp.bgcol_;
    im_ = temp.im_;
    // POST: this and temp are equal

    temp.im_ = 0;
    // POST: temp won't deallocate the image

    return *this;
}

gd_image::~gd_image() {
    if (0 != im_) { // Note it can be set to 0 (only) internally
        ::gdImageDestroy(im_);
    }
}

gd_image gd_image::vmerge(const gd_image & top, const gd_image & bottom) {
    const int w = std::max<int>(top.w_, bottom.w_);
    const int h = top.h_ + bottom.h_;

    gd_image rg(w, h, top.bgcol_);

    ::gdImageCopy(rg.im_, top.im_, 0, 0, 0, 0, top.w_, top.h_);
    ::gdImageCopy(rg.im_, bottom.im_, 0, top.h_+1, 0, 0, bottom.w_, bottom.h_);

    return rg;
}

gd_image &  gd_image::flood(int colidx) {
    flood(im_, colidx);
    return *this;
}

// static
void gd_image::flood(::gdImagePtr & im, int colidx) {
    ::gdImageFill(im, 0,0, colidx);
}

gd_image & gd_image::resize(int w, int h) {
    if (w == w_ && h == h_) { return *this; };

    ::gdImagePtr im = ::gdImageCreateTrueColor(w, h);
    if (0 == im) {
        // TODO: Do something
    }
    // Ensure the bg is the same
    flood(im, ::gdImageColorAllocate(im, bgcol_.r, bgcol_.g, bgcol_.b));
    ::gdImageCopyResized(im, im_, 0, 0, 0, 0, w, h, w_, h_);

    // Can safely replace the contents:
    assert( im != 0 );
    ::gdImageDestroy(im_);

    im_ = im;
    w_ = w;
    h_ = h;

    return *this;
}

gd_image & gd_image::expand(int w, int h) {
    // PRE:
    assert( w >= w_ ); assert( h >= h_ );

    if (w == w_ && h == h_) { return *this; }

    ::gdImagePtr im = ::gdImageCreateTrueColor(w, h);
    if (0 == im) {
        // TODO: Do something
    }
    // Ensure the bg is the same
    flood(im, ::gdImageColorAllocate(im, bgcol_.r, bgcol_.g, bgcol_.b) );

    ::gdImageCopy(im, im_, 0, 0, 0, 0, w_, h_);

    assert( im != 0 );

    im_ = im;
    w_ = w;
    h_ = h;

    return *this;
}

void gd_image::line(int x0, int y0, int x, int y, int colidx) {
    ::gdImageLine(im_, x0, y0, x, y, colidx);
}

void gd_image::rectangle(int x1, int y1, int x2, int y2, int colidx, bool filled) {
    if (filled) {
        ::gdImageFilledRectangle(im_, x1, y1, x2, y2, colidx);
    }
    else {
        ::gdImageRectangle(im_, x1, y1, x2, y2, colidx);
    }
}

void gd_image::text(int x, int y, int colidx, const std::string & text) {
    // Ugly:
    unsigned char * str = reinterpret_cast<unsigned char *>( const_cast<char *>(text.c_str()) );
    ::gdImageString(im_, ::gdFontGetLarge(), x, y, str, colidx);
}

int gd_image::get_font_height() const {
    return ::gdFontGetLarge()->h;
}

int gd_image::get_font_width() const {
    return ::gdFontGetLarge()->w;
}

int gd_image::allocate_colour(int r, int g, int b) {
    assert( r < 256 );
    assert( g < 256 );
    assert( b < 256 );
    //return ::gdImageColorAllocate(im_, r, g, b);
    // ColorExact will retrieve the color if allocated or allocate a new
    // one otherwise. ColorAllocate will allocate a new one always.
    // Not sure if it matters at all with true color...
    return ::gdImageColorExact(im_, r, g, b);
}

void gd_image::write(std::ofstream & target) const {
    ::gdSink sink;
    sink.context = &target;
    sink.sink = sink_callback;
    ::gdImagePngToSink(im_, &sink);
}

// static
int gd_image::sink_callback(void * context, const char * buffer, int len) {
    assert( 0 != context );
    std::ofstream & o = * reinterpret_cast<std::ofstream *>(context);

    // return -1 on error (see gd.h), length of bytes written on success

    if (!o.good()) { return -1; }

    o.write(buffer, len);

    // write sets badbit on failure
    if (o.bad()) { return -1; }

    return len;
}

std::ofstream & operator<<(std::ofstream & ofs, const gd_image & im) {
    im.write(ofs);
    return ofs;
}

} // namespace mp3plot

#endif // GD / GRAPHICS
// vim:set ts=4 et ai:


