/*
 * mp3plot - Bitrate analysis tool
 *
 * Copyright (C) 2007, 2009 Toni Corvera
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

// $Id: gd_image.h 1208 2009-04-24 01:25:31Z  $

#ifndef MP3PLOT_GD_IMAGE_H
#define MP3PLOT_GD_IMAGE_H

#include "../mp3_config.h"

#if defined(HAVE_GD) && !defined(DISABLE_GRAPHICS) && !defined(DISABLE_GD)

#include <map>
#include <fstream>
#include <gd.h>

#include "../errors.h"
#include "colour.h"

namespace mp3plot {

    using net_outlyer::colour_type;
    using net_outlyer::rgb_type;

    /*!
     * \brief Wrapper around GD following the RAII pattern.
     * \note This isn't meant to be a general use GD wrapper, only the
     * methods required in the context of this program are implemented.
     * \see http://www.boutell.com/gd/manual2.0.33.html
     */
    class gd_image {
    public:
        typedef colour_type<int,255> rgb24b_colour;
        // Colour type used by the gd library
        typedef int                  gd_colour_t;

        /*!
         * \brief Constructor.
         * \param[in] w Image width
         * \param[in] h Image Height
         * \param[in] bgcol Background colour
         * \sa resize(int, int)
         * \throw e_error If initialisation failed
         */
        gd_image(int w, int h, const rgb24b_colour & bgcol) throw (e_error);

        /*!
         * \brief Copy constructor.
         * The new image will be a full copy of the source.
         * \throw e_error If initialisation failed
         */
        gd_image(const gd_image &) throw (e_error);

        /*!
         * \brief Assignment operator.
         * The new image will be a full copy of the source.
         *
         */
        gd_image & operator=(const gd_image &) throw (e_error);

        /*!
         * \brief Destructor. Frees the allocated memory.
         * \note The image won't be automatically dumped in a file.
         * \sa write(std::ofstream &)
         */
        ~gd_image();

        //! Flood the image from 0,0
        gd_image & flood(int colidx);

        /*!
         * \brief Image Join operator.
         */
        static gd_image vmerge(const gd_image &, const gd_image &);

        /*!
         * \brief Resizes the image.
         */
        gd_image & resize(int, int);

        /*!
         * \brief Resizes the "canvas".
         * \param[in] w Target width
         * \param[in] h Target height
         * \pre w and h are at equal or bigger than the current dimensions.
         */
        gd_image & expand(int, int);

        /*!
         * \brief Draws a line.
         * \param[in] x0 Source X coordinate
         * \param[in] y0 Source Y coordinate
         * \param[in] x Target X coordinate
         * \param[in] y Target Y coordinate
         * \param[in] colour Colour for the line
         */
        void line(int x0, int y0, int x, int y, int colour);

        /*!
         * \brief Draws a rectangle of the specified colour.
         * \param[in] x1 Top-left corner's X
         * \param[in] y1 Top-left corner's Y
         * \param[in] x2 Bottom-right corner's X
         * \param[in] y2 Bottom-right corner's Y
         * \param[in] colour Colour of the rectangle
         * \param[in] filled When true the rectangle will be filled, when false, only
         *           the outline will be drawn.
         */
        void rectangle(int x1, int y1, int x2, int y2, int colour, bool filled=false);

        /*!
         * \brief Allocates a colour in the image.
         * The colour is implemented in the (floating point) range [0..1]
         * \note gd works with ints so
         *       allocate_colour( const colour_type<int,255>& ) is
         *       preferred as the extra precission will be lost anyway.
         * \sa allocate_colour( const colour_type<int,255> & c )
         */
        int allocate_colour(const rgb_type & c) {
            return allocate_colour( c.convert<int,255>() );
        }

        /*!
         * \brief Allocates a colour in the image.
         * \sa allocate_colour(const colour_type<double,1> & c)
         */
        int allocate_colour(const rgb24b_colour & c) {
            return allocate_colour(c.r, c.g, c.b);
        }

        /*!
         * \brief Writes the image as a PNG to a file.
         * \sa operator<<(std::ofstream &, const gd_image &)
         */
        void write(std::ofstream &) const;

        /*!
         * \brief Draw text to the image
         * \param[in] x Top-left, x coordinate
         * \param[in] y Top-left, y coordinate
         * \param[in] colour Text colour
         * \paran[in] text Text to draw
         */
        void text(int x, int y, int colour, const std::string & text);

        int get_font_height() const;
        int get_font_width() const;

    protected:
        /*!
         * \brief Allocates a colour in the image.
         * This is the lower level allocation, more in line with the raw
         * gd library, the abstracted methods are preferred, though.
         * \remarks In True-color images, the first allocation will define
         * the background colour.
         * \remarks Colour components are in the range [0..255]
         * \param[in] r Red component
         * \param[in] g Green component
         * \param[in] b Blue component
         * \sa allocate_colour(const colour_type<double,1> &)
         * \sa allocate_colour(const colour_type<int,255> &)
         */
        int allocate_colour(int r, int g, int b);

        static void flood(::gdImagePtr & im, int colidx);

        /*!
         * \brief Function called when writing from memory to a file.
         */
        static int sink_callback(void * context, const char * buffer, int len);

    private:
        ::gdImagePtr im_; //!< The internal gd image
        int w_, //!< Width
        h_; //!< Height
        rgb24b_colour bgcol_;
    }; // class gd_image

    // Forward declaration
    /*!
    * \brief File Output operator for the gd_image.
    * \remarks Allows writing an image with:
    * \code
    * ofstream ofs;
    * gd_image im;
    * ...
    * ofs << im;
    * \endcode
    */
    std::ofstream & operator<<(std::ofstream &, const gd_image &);

    namespace gdlib {
        typedef gd_image image;
        typedef gd_image::gd_colour_t   colour;
        typedef gd_image::rgb24b_colour rgb24b_colour;
    }

} // namespace mp3plot

#endif // GD / GRAPHICS

#endif // MP3PLOT_GD_IMAGE_H
// vim:set ts=4 et ai:

