/*
 * mp3plot - Bitrate analysis tool
 *
 * Copyright (C) 2007, 2009 Toni Corvera
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

// $Id: gd_plotter.cc 1213 2009-04-24 20:18:17Z  $

#include "../mp3_config.h"                 // Defines HAVE_GD / DISABLE_GD / DISABLE_GRAPHICS

#if defined(HAVE_GD) && !defined(DISABLE_GRAPHICS) && !defined(DISABLE_GD)

#include <string>
#include <cassert>
#include <fstream>

#include "../common.h"
#include "../utils/logger.h"
#include "../mp3/bitrate.h"
#include "../mp3/frame.h"
#include "colour.h"
#include "gd_plotter.h"
#include "gd_image.h"

namespace {

    using mp3plot::numframes_t;
    using mp3plot::gd_image;
    using mp3plot::graphics_lib_interface;

    namespace gd = mp3plot::gdlib;

    // This is pretty much redundant...
     struct gd_interface : public graphics_lib_interface<gd::image, gd::colour> {
        gd_interface()
            : cached_bordercol(0)
        {}

        virtual ~gd_interface() {
            delete cached_bordercol;
        }
        virtual image_type create_image(int w, int h) const {
            return gd_image(w, h, gd::rgb24b_colour(0xff));
        }

        virtual void resize(image_type & im, int w, int h) const {
            im.resize(w, h);
        }

        virtual image_type vertical_merge(const image_type& top, const image_type& bottom) const {
            return gd_image::vmerge(top, bottom);
        }

        virtual native_colour allocate_colour(image_type& im,
                                              const common_colour& c) const {
            return im.allocate_colour(c);
        }

        virtual void draw_line(image_type & im,
                               int x0, int y0, int x1, int y1,
                               const native_colour & c) const
        {
            im.line(x0, y0, x1, y1, c);
        }

        virtual void draw_rectangle(image_type& im,
                                    int x0, int y0, int x1, int y1,
                                    const native_colour& c) const
        {
            // Order is not important:
            //assert( x0 < x1 );
            //assert( y0 < y1 );
            // Empty rectangle:
            if (x0 == x1 && y0 == y1) { return; }
            // "negative" rectangle
            if (x0+1 >= x1-1 || y0+1 >= y1-1) { return; }

            if (0 != cached_bordercol) {
                if (cached_bordercol->first != &im) {
                    delete cached_bordercol;
                    cached_bordercol = 0;
                }
            }
            if (0 == cached_bordercol) {
                const int col = im.allocate_colour(gd::rgb24b_colour(0));
                cached_bordercol = new std::pair<gd_image*,int>(&im, col);
            }
            assert( 0 != cached_bordercol );
            assert( &im == cached_bordercol->first );
            const int black = cached_bordercol->second;
            im.rectangle(x0, y0, x1, y1, black, false);
            im.rectangle(x0+1, y0+1, x1-1, y1-1, c, true);
        }

        // In GD, the coordinates are those of the top-left corner
        virtual void draw_text(image_type& im,
                               int x, int y,
                               const native_colour & c,
                               const std::string & label) const
        {
            const int tw = im.get_font_width() * label.length();
            im.text(x-(tw/2), y-(im.get_font_height()/2), c, label);
        }
    private:
        mutable std::pair<gd_image*, int> * cached_bordercol;
    };

} // anonymous namespace

namespace mp3plot {

void gd_plotter::plot(const mp3file & file, const plot_params& params) const
    throw (e_error)
{
    assert( !params.output_filename.empty() );

    using namespace std;
    net_outlyer::logger::os() << "Using the GD output module\n";

    try {
        const gd_image im = image_plotter::plot<gdlib::image, gdlib::colour>(file.get_frames(), gd_interface(), params);
        std::ofstream ofs( params.output_filename.c_str(), ios::out|ios::binary );
        ofs << im;
        ofs.close();
    }
    catch (const e_error & e) {
        // Only gd_image() throws
        throw e_error("Failed to create images in memory.", EX_UNAVAILABLE);
    }
} // gd_plotter::plot()


} // namespace mp3plot

#endif // DISABLE_GRAPHICS, DISABLE_GD

// vim:set ts=4 et ai:

