/*
 * mp3plot - Bitrate analysis tool
 *
 * Copyright (C) 2007, 2009 Toni Corvera
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

// $Id: magickpp_plotter.cc 1216 2009-04-25 13:48:49Z  $

#include "magickpp_plotter.h"   // HAVE_MAGICKPP / DISABLE_MAGICK / DISABLE_GRAPHICS

#if defined(HAVE_MAGICKPP) && !defined(DISABLE_GRAPHICS) && !defined(DISABLE_MAGICK)

#include <string>
#include <list>
#include <cassert>
#include <Magick++.h>

#include "../utils/logger.h"
#include "../mp3/bitrate.h"
#include "../mp3/frame.h"
#include "../common.h"
#include "graphics_lib_interface.h"
#include "colour.h"

// Magick::Image Reference: <http://www.graphicsmagick.org/Magick++/Image.html>

namespace { // anon
    using net_outlyer::rgb_type;

    std::string bestFont;
    bool bestFontSet = false;
    // Find the best possible font TODO: Is there a native way to check font availability
    void find_best_font() {
        Magick::Image im(Magick::Geometry(100,100),Magick::ColorGray(0));
        Magick::DrawableText t(0,0,"A");

        try {
            // XXX: Is fixed granted?.
            // (Is it an alias for Helvetica ($convert -list font)?)
            // Setting an unusable/unavailable font will throw WarningType
            // upon draw()
            im.font("fixed");
            im.draw(t);
        }
        catch (const Magick::WarningType& e) {
            try {
                im.font("Cou0rier");
                im.draw(t);
            }
            catch (const Magick::WarningType& e) {
                im.font("");
            }
        }
        bestFont = im.font();
        bestFontSet = true;
    }

    struct magickpp_wrapper : public mp3plot::graphics_lib_interface<Magick::Image, Magick::Color> {
        virtual Magick::Color allocate_colour(image_type& im, const rgb_type & c) const {
            return Magick::ColorRGB(c.r, c.g, c.b);
        }
        virtual image_type create_image(int w, int h) const {
            if (!bestFontSet) {
                find_best_font();
            }
            Magick::Image im(Magick::Geometry(w,h), Magick::ColorGray(1.0f));
            im.font(bestFont); // see find_best_font() for details
            im.fontPointsize(14);
            //im.antiAlias(false); // font-size is pretty small
            //im.strokeAntiAlias(false);
            return im;
        }
        virtual void draw_line(image_type & im, int x0, int y0, int x1, int y1, const native_colour& c) const {
            im.strokeColor(c);
            im.strokeWidth(1);
            im.draw( Magick::DrawableLine(x0, y0, x1, y1) );
        }
        virtual void draw_rectangle(image_type& im, int x0, int y0, int x1, int y1, const native_colour& c) const {
            // XXX: strokeWidth(0) doesn't remove the border (???)
            im.strokeWidth(1);
            im.strokeColor(Magick::ColorGray(0));
            im.fillColor(c);
            im.lineWidth(0);
            im.draw( Magick::DrawableRectangle(x0,y0,x1,y1) );
        }
        virtual void resize(image_type& im, int w, int h) const {
            Magick::Geometry g = Magick::Geometry(w, h);
            g.aspect(true); // ImageMagick's "!"
            im.sample(g);
        }
        // In Magick++, the coordinates are those of the top-left corner
        virtual void draw_text(image_type& im,
                               int x, int y,
                               const native_colour& c,
                               const std::string & label) const
        {
            im.strokeWidth(1);
            //im.fillColor(c);
            Magick::TypeMetric tm;
            im.fontTypeMetrics(label, &tm);
            const int xoff = static_cast<int>(tm.textWidth() / 2);
            const int yoff = static_cast<int>(tm.textHeight() / 3);
            // Font and pointsize are set on creation, see create_image()
            im.draw( Magick::DrawableText(x-xoff, y+yoff, label) );
        }
        virtual image_type vertical_merge(const image_type& top, const image_type& bottom) const {
            // app will hold the appended images
            image_type app;
            std::list<image_type> L;
            L.push_back(top);
            L.push_back(bottom);
            Magick::appendImages(&app, L.begin(), L.end(), true); // true => append vertically
            return app;
        }
    };
} // anon namespace

namespace mp3plot {

void magickpp_plotter::plot(const mp3file & file, const plot_params& params) const
    throw (e_error)
{
    assert( !params.output_filename.empty() );

    using namespace Magick;

    net_outlyer::logger::os() << "Using the Magick++ output module\n";

    const frame_collection & frames = file.get_frames();

    Magick::Image im = image_plotter::plot<Image,Color>(frames, magickpp_wrapper(), params);
    im.magick("png"); // Force PNG output even if filename isn't a png file
    im.write(params.output_filename);
} // magickpp_plotter::plot()

} // namespace mp3plot

#endif // DISABLE_GRAPHICS, DISABLE_MAGICK

// vim:set ts=4 et ai:

