/*
 * mp3plot - Bitrate analysis tool
 *
 * Copyright (C) 2007, 2009 Toni Corvera
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

// $Id: plotter_features.cc 1090 2009-04-01 20:15:14Z  $

#include "plotter_features.h"

#include <iostream>

namespace mp3plot {

// Constants
// These are used in the bitset so their binary representation must not overlap
const plotter_features::constant_type
                    plotter_features::TEXTUAL   = 0x01U,
                    plotter_features::GRAPHICAL = 0x02U,
                    plotter_features::MARKUP    = 0x04U,
                    plotter_features::PRINTED   = 0x08U,
                    plotter_features::FILE      = 0x10U;

plotter_features::plotter_features(const bitset_type & b)
    : bits(b)
{}

const plotter_features plotter_features::textual() {
    return plotter_features(TEXTUAL);
}

const plotter_features plotter_features::graphical() {
    return plotter_features(GRAPHICAL);
}

const plotter_features plotter_features::printed() {
    return plotter_features(PRINTED);
}

const plotter_features plotter_features::file() {
    return plotter_features(FILE);
}

bool plotter_features::operator==(const plotter_features & to) const {
    if (this == &to) {
        return true;
    }

    return to.bits == bits;
}

bool plotter_features::operator>(const plotter_features & p) const {
    return (p != *this) && (bits & p.bits).any();
}

bool plotter_features::operator>=(const plotter_features & p) const {
    return (p == *this) || (bits & p.bits).any();
}


plotter_features plotter_features::operator|(const plotter_features & p) const {
    return plotter_features(bits | p.bits);
}

plotter_features plotter_features::operator&(const plotter_features & p) const {
    return plotter_features(bits & p.bits);
}

plotter_features plotter_features::operator^(const plotter_features & p) const {
    return plotter_features(bits ^ p.bits);
}

plotter_features plotter_features::operator!() const {
    return plotter_features( (bitset_type(bits)).flip() );
}

// [C++] local visibility
namespace {
    inline const char * YESNO(bool v) {
        return ( v ? "yes" : "no" );
    }
}

// output operator
std::ostream & operator<<(std::ostream & os, const plotter_features & pf) {
    typedef plotter_features::bitset_type bst;

    const bst t = bst(plotter_features::TEXTUAL),
              g = bst(plotter_features::GRAPHICAL),
              m = bst(plotter_features::MARKUP),
              p = bst(plotter_features::PRINTED),
              f = bst(plotter_features::FILE);

    return os <<
    "is_textual = "     << YESNO( (pf.bits & t).any() ) <<
    "\nis_graphical = " << YESNO( (pf.bits & g).any() ) <<
    "\nis_markup = "    << YESNO( (pf.bits & m).any() ) <<
    "\nis_printed = "   << YESNO( (pf.bits & p).any() ) <<
    "\nis_file = "      << YESNO( (pf.bits & f).any() ) <<
    "\n";
}

std::ostream & operator<<(std::ostream & os, const plotter_description & pd) {
    os << pd.name() << ": ";

    bool ws = false, pws = false;
    if ((ws = (pd.feats_ > plotter_features::textual()) )) {
        os << "textual";
        pws = ws;
    }
    if ((ws = (pd.feats_ > plotter_features::graphical()) )) {
        os << (pws ? ", " : "") << "graphical";
        pws = ws;
    }

    if ((ws = (pd.feats_ > plotter_features::printed()) )) {
        os << (pws ? ", " : "") << "printed";
        pws = ws;
    }
    if ((ws = (pd.feats_ > plotter_features::file()) )) {
        os << (pws ? ", " : "") << "file output";
        pws = ws;
    }

    return os;
}

} // namespace mp3plot

// vim:set ts=4 et ai:

