/*
 * mp3plot - Bitrate analysis tool
 *
 * Copyright (C) 2007, 2009 Toni Corvera
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

// $Id: plotter_features.h 1106 2009-04-07 00:07:31Z  $

#ifndef MP3PLOT_PLOTTER_FEATURES_H
#define MP3PLOT_PLOTTER_FEATURES_H

#include <ostream>
#include <bitset>
#include <climits>

namespace mp3plot {

    using std::ostream;

    // MSVC compat
    class plotter_features;
    ostream & operator<<(ostream&, const plotter_features&);

    /*!
     * \brief Abstraction of the description of the capabilities of a plotter
     * \see bitrate_plotter
     * \remarks This class isn't meant to be instantiated, you should use it
     *          with the \em static \em constructors
     * \see textual(), graphical(), printed(), file()
     * \remarks Bitwise operators have boolean algebra consistency, e.g.:
     * \code
     * plotter_features p, q;
     * ...
     * p == !!p ;            // true
     * p ^ !p == p | !p ;    // tue
     * ( q | p ) & !q == p ; // true
     * !(p | q) == !p & !q;  // true
     * \endcode
     * \see operator !
     * \see operator &
     * \see operator |
     * \see operator ^
     */
    class plotter_features {
    public:

        // output operator, for debugging purposes
        friend std::ostream & mp3plot::operator<<(std::ostream &, const plotter_features &);

        //! Returns a text-only description
        static const plotter_features textual();

        //! Returns a graphical-only description
        static const plotter_features graphical();

        /*!
         * \brief Returns a printed-output plotter.
         * \remarks This is not exclusive with a file-output plotter.
         * \see file()
         */
        static const plotter_features printed();

        /*!
         * \brief Returns a file-output plotter.
         * \remarks This is not exclusive with a printed-output plotter.
         * \see printed()
         */
        static const plotter_features file();

        //! Equivalence operator
        bool operator==(const plotter_features & to) const;

        //! Negated equivalence operator
        bool operator!=(const plotter_features & p) const { return !(p==(*this)); }

        /*!
         * \brief Inclusion operator.
         * A > B if A != B and B is a subset of the features of A.
         *
         */
        bool operator>(const plotter_features & p) const;

        /*!
         * \brief Inclusion/equivalence operator.
         * A >= B if A == B or B is a subset of the features of A.
         */
        bool operator>=(const plotter_features & p) const;

        /*!
         * \brief Bitwise OR operator.
         * The result will have the capabilities of both plotter_features
         */
        plotter_features operator|(const plotter_features & p) const;

        /*!
         * \brief Bitwise AND operator.
         * The result will have the common capabilities
         */
        plotter_features operator&(const plotter_features & p) const;

        /*!
         * \brief Addition operator, semantically equal to bitwise OR.
         * \see operator|()
         * \deprecated operator|() is preferred, operator+() is only provided
         *             because it makes some sense semantically and provides
         *             some tighter syntax when operator precedence makes a
         *             difference. e.g., it might avoid the use of parentheses
         *             if the programmer so chooses:
         * \code
         * std::cout << plotter_features::textual() + plotter_features::printed();
         * \endcode
         *             Instead of
         * \code
         * std::cout << ( plotter_features::textual() | plotter_features::printed() );
         * \endcode
         */
        plotter_features operator+(const plotter_features & p) const { return (*this)|p; }

        /*!
         * \brief Bitwise XOR operator.
         * The result will have the features in one of the operands but not
         * in both.
         */
        plotter_features operator^(const plotter_features & p) const;

        /*!
         * \brief Logical NOT.
         * The will have the exact oposite features.
         */
        plotter_features operator!() const;

        /*!
         * \brief Default, empty constructor.
         * \deprecated Only use to declare variables to be assigned later.
         */
        plotter_features()
            : bits(0)
        {}
    private:
        /*!
         * \brief Type for the symbolic constants.
         * \remarks Should be <= than the type of bitset.to_ulong() (i.e.
         * unsigned long).
         */
        typedef unsigned int constant_type;
        /*!
         * \brief Type for the internally used bitset.
         * \remarks About the size of the bitset:
         * Should be capable of holding any \em contant_type
         * \em Remember: A byte in C++ is *at least (read: might be more)* 8 bits.
         * CHAR_BIT (defined in &lt;climits&gt;) contains the size of a \em C++ \em byte.
         * \see constant_type
         */
        typedef std::bitset< CHAR_BIT * sizeof(constant_type) > bitset_type;

        //! Internal constructor
        explicit plotter_features(const bitset_type &);

        bitset_type bits;

        static const constant_type // Type should match bitset.to_ulong() !
                TEXTUAL,   //!< The plotter produces text as output
                GRAPHICAL, //!< The plotter produces an image as output
               /*!
                * \brief The plotter produces markup text as output.
                * \note Currently unused
                * \remarks is_markup should encompass is_textual
                */
                MARKUP,
                PRINTED,   //!< The plotter outputs to stdout
                FILE       //!< The plotter outputs to a file
        ; // End of Constants
    };

    class plotter_description {
    public:
        plotter_description(const std::string & name, const plotter_features & feats)
            : name_(name), feats_(feats)
        {}
        bool matches_features(const plotter_features & pf) const { return pf == feats_; }
        bool includes_features(const plotter_features & pf) const { return feats_ >= pf; }
        const std::string & name() const { return name_; }

        friend std::ostream & operator<<(std::ostream &, const plotter_description &);
    private:
        const std::string name_;
        const plotter_features feats_;
    };

    std::ostream & operator<<(std::ostream &, const plotter_description &);

    // Forward declaration for documentation purposes
    /*!
     * \brief Output operator, for debugging purposes.
     * Prints the features one per line, e.g.:
     * \code
     * std::cout << ( plotter_features::textual() | plotter_features::file() );
     * \endcode
     * would produce
     * \code
     * is_textual = yes
     * is_graphical = no
     * is_markup = no
     * is_printed = no
     * is_file = yes
     * \endcode
     * \see operator |
     * \remarks The last line is ended in a newline character
     */
    std::ostream & operator<<(std::ostream &, const plotter_features &);

} // namespace mp3plot

#endif // MP3PLOT_PLOTTER_FEATURES_H
// vim:set ts=4 et ai:

