/*
 * mp3plot - Bitrate analysis tool
 *
 * Copyright (C) 2007, 2009 Toni Corvera
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

// $Id: spinner.cc 1215 2009-04-25 12:47:38Z  $

#include "spinner.h"

#include <iostream>
#include <cassert>
#include <cstring> // strlen
#include <cstdlib> // exit

// FIXME!!!: I've *once so far* found a threadlock while running the
// new, Boost.Thread-based, code. Should look into it

#define SYMS_DEF "|/-\\"

#include <boost/version.hpp> // The use of boost conditions changed at some point...
#include <boost/ref.hpp>
#include <boost/thread.hpp>

#if BOOST_VERSION < 103401
    // Conditions were introduced in 1.34.1, lower versions aren't supported
#   error At least boost 1.34.1 is required
#elif BOOST_VERSION < 103500
    // the locking code changed in 1.35.0. Lower versions code:
    typedef boost::condition                 condition_type;
    typedef boost::mutex::scoped_lock        lock_type;
#else
    typedef boost::condition_variable        condition_type;
    typedef boost::unique_lock<boost::mutex> lock_type;
#endif

using namespace std;

namespace {
    const size_t LEN=::strlen(SYMS_DEF);
    const char * SYMS_ = SYMS_DEF;

    condition_type cond;
    boost::mutex lmutex; // not really used, but condition requires it

    /**
     * \brief Functor with the actual spinner code
     */
    struct impl {
        void operator()() {
            must_stop = false;
            idx = 0;

            while (!must_stop) {
                lock_type lock(lmutex);
                cond.wait(lock); // wait until notify_*
                cerr << "\r[" << SYMS_[static_cast<int>(idx)] << "]";
                idx = ( idx + 1 ) % LEN;

                if(must_stop) break;
            }
            // Cleanup output
            cerr << flush << "\r" << flush;
        }
        volatile bool must_stop;
        int idx;
    };
}

namespace net_outlyer {

const char * spinner::SYMS = SYMS_DEF;

impl the_impl;

void spinner::start() {
    lock_type lock(lmutex);
    // Note the red is VITAL, boost::thread takes makes a private copy of the
    // object. Due to the way the program is written that duplicated the
    // mutex/condition code leading to assertion failures when deallocating
    // them
    // <http://www.boost.org/doc/libs/1_38_0/doc/html/thread/thread_management.html>
    boost::thread t(boost::ref(the_impl));
    // On exit from start(), t will become a detached thread
}

// This is called from the parent thread, it writes a byte on the shared pipe
// deprecated
void spinner::step() {
    (*this)++;
}

// Can't define with a return type, it stopped working

void spinner::operator++(int) {
    cond.notify_one();
}

void spinner::operator++() {
    cond.notify_one();
}

// This is called from the parent thread, instructs the child thread to stop spinning
// For some reason the end of the loop isn't reached (see impl's details)
void spinner::stop() {
    the_impl.must_stop = true;
    cond.notify_one(); // Otherwise won't exit the loop
}

spinner::~spinner() {
    stop();
}


} // namespace net_outlyer

// vim:set ts=4 et ai:
