/*
 * Copyright (c) 2009, Toni Corvera. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

// $Id: byteswap.hxx 1344 2009-06-22 01:11:09Z toni $

/*!
 * \file
 * \brief Re-implementation of \c ntohl/ntohs/htonl/htons
 * \remarks \c WORDS_BIGENDIAN is used to check for endianness
 */
#ifndef ONET_BYTESWAP_HXX
#define ONET_BYTESWAP_HXX

#include <stdint.h>

#if 0 // This would be the conditional inclusion needed to use the standard version
#   if HAVE_NTOHL_IN_WINSOCK2_H
#       include <winsock2.h>
#   elif HAVE_NTOHL_IN_ARPA_INET_H
#       include <arpa/inet.h>
#   elif HAVE_NTOHL_IN_NETINET_IN_H
#       include <netinet/in.h>
#   elif HAVE_NTOHL_IN_SYS_TYPES_H
#       include <sys/types.h>
#   else
#       error "no known ntohl definition found"
#   endif
#endif

namespace net_outlyer {

namespace {

// Byte-swapping implementation, use byteswap or ntohl and pals instead

// General implementation
// Definition as a struct allows to specialise on the important types
// <http://yosefk.com/c++fqa/templates.html#fqa-35.7>
template<typename T, int S> struct SwapBytesI {
    static T impl(const T& v) {
        T r;
        const char *pg = reinterpret_cast<const char*> (&v);
        char *pp = reinterpret_cast<char*> (&r) + sizeof (T) - 1;

        for (int off = 0; off < S; ++off) {
            *pp-- = *pg++;
        }
    }
};
// 4 bytes implementation
template<typename T> struct SwapBytesI<T, 4> {
    static T impl(const T& v) {
        return (v&0xFF000000)>>24 | (v&0x00FF0000)>>8 | (v&0x0000FF00)<<8 | (v&0x000000FF)<<24;
    }
};
// 2 bytes implementation
template<typename T> struct SwapBytesI<T, 2> {
    static T impl(const T& v) {
        return (v&0xFF00)>>8 | (v&0x00FF)<<8;
    }
};

} // anon ns

/*!
 * \brief Swap all bytes in a copy of the input
 */
template<typename T> T byteswap(const T& t) {
    return SwapBytesI<T, sizeof(T)>::impl(t);
}

//////////////////////////////////////////////////////
// Forward declarations, for documentation purposes //
//////////////////////////////////////////////////////

/*!
 * \brief Swap bytes in a 32 bytes value, from big-endian to the system used
 *        in this host (i.e. do nothing if the system is big-endian)
 */
template<typename T> T inline btoh32(const T&);
/*!
 * \brief Swap bytes in a 16 bytes value, from big-endian to the system used
 *        in this host (i.e. do nothing if the system is big-endian)
 */
template<typename T> T inline btoh16(const T&);
/*!
 * \brief Swap bytes in a 32 bytes value, from the system used in this host to
 *        big-endian (i.e. do nothing if the system is big-endian)
 */
template<typename T> T inline htob32(const T&);
/*!
 * \brief Swap bytes in a 16 bytes value, from the system used in this host to
 *        big-endian (i.e. do nothing if the system is big-endian)
 */
template<typename T> T inline htob16(const T&);

// Defining these as ntohl and pals would cause problems if the system's
// versions are defined as macros
// Instead, the n (for network) has been replaced by b (for big-endian),
// and the suffixes (l for long, s for short) have been replaced by 32 and 16
#ifdef WORDS_BIGENDIAN
template<typename T> T inline btoh32(const T& v) {
    return v;
}

template<typename T> T inline btoh16(const T& v) {
    return v;
}
#else // Little endian

/*!
 * \brief Swap bytes in a 32 bytes value, from big-endian to the system used
 *        in this host (i.e. do nothing if it's big-endian)
 */
template<typename T> T inline btoh32(const T& v) {
    const uint32_t V = static_cast<uint32_t>(v);
    const uint32_t S = byteswap(V);
    return static_cast<T>(S);
    // Problematic: return static_cast<T>(byteswap(static_cast<uint32_t>(v)));
}

template<typename T> T inline btoh16(const T& v) {
    const uint16_t V = static_cast<uint16_t>(v);
    const uint16_t S = byteswap(V);
    return static_cast<T>(S);
    //return static_cast<T>(byteswap(static_cast<uint16_t>(v)));
}
#endif // Endianness

template<typename T> T inline htob32(const T& v) {
    return btoh32(v);
}

template<typename T> T inline htob16(const T& v) {
    return btoh16(v);
}

// TODO: Add XtoX64

} // ns net_outlyer

#endif // ONET_BYTESWAP_HXX
// vim:set ts=4 et ai:

