/*
 * Copyright (c) 2008, 2009 Toni Corvera. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

// $Id: constraints.hxx 1344 2009-06-22 01:11:09Z toni $

/*!
 * \file
 * \brief Common C++ constraints.
 */
#ifndef ONET_CONTRAINTS_HXX
#define ONET_CONTRAINTS_HXX

#include <iostream>

/*!
 * \def ONET_CT_HXX_UNUSED
 * Disables warning about an unused variable in compilers supporting it
 */
#ifdef __GNUC__
#   define ONET_CT_HXX_UNUSED __attribute__((unused))
#else
#   define ONET_CT_HXX_UNUSED // Nothing
#endif

namespace net_outlyer {

namespace constraints {

/*!
 * \struct converts_to
 * \brief Class-conversion constraint
 * \param T Type to check
 * \param Base Target type
 * \sa Based on http://www.research.att.com/~bs/bs_faq2.html#constraints
 */
template<class T, class Base> struct converts_to {
    static void constraints(T* p) { ONET_CT_HXX_UNUSED Base* pb = p; }
    converts_to() { ONET_CT_HXX_UNUSED void(*p)(T*) = constraints; }
};

/*!
 * \struct derived_from
 * \brief Class-hierarchy constraint
 * \param T Type to check
 * \param Base Ancestor type
 * \remarks It actually checks for conversion and not only derivation
 * \sa Based on http://www.research.att.com/~bs/bs_faq2.html#constraints
 */ // Currently an alias for converts_to
template<class T, class Base> struct derived_from : converts_to<T, Base> {
};

/*!
 * \struct static_casts_to
 * \brief Class-conversion constraint.
 * \param TFrom Type to check
 * \param TTo Target type
 *
 * Checks if a type \p TFrom casts \b statically to another type \p TTo
 */
template<class TFrom, class TTo> struct static_casts_to {
    static void constraint(const TFrom & a) {
        ONET_CT_HXX_UNUSED TTo N = static_cast<TTo>(a);
    }
    static_casts_to() {
        ONET_CT_HXX_UNUSED void (*ct)(const TFrom&) = constraint;
    }
};

/*!
 * \struct is_printable
 * \brief Checks if a type can be sent to a \c std::stream, i.e. if there's an
 *        output operator available for the type.
 * \param T Type to check
 */
template<typename T> struct is_printable {
    static void constraint(const T& t) {
        std::cerr << t;
    }
    is_printable() {
        ONET_CT_HXX_UNUSED void (*ct)(const T&) = constraint;
    }
};

} // ns constraints

} // ns outlyer


#endif // ONET_CONTRAINTS_HXX
// vim:set ts=4 et ai: //
