/*
 * Copyright (c) 2007-2009, Toni Corvera. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

// $Id: named_fstream.hxx 1344 2009-06-22 01:11:09Z toni $

/*!
 * \file
 * \brief Filestream classes retaining filename.
 */
#ifndef ONET_NAMED_FSTREAM_HXX
#define ONET_NAMED_FSTREAM_HXX

#include <fstream>
#include <string>

#include "named_file.hxx"
#include "constraints.hxx"

namespace net_outlyer {

namespace file {

    using std::ofstream;
    using std::ifstream;
    using std::fstream;
    using std::string;
    using std::ios_base;

    /*!
     * \brief Abstract base class for named std::ios
     * \pre \p stream_type derives from \c std::ios
     */
    template<typename stream_type> struct named_ios : public named_file {
    protected:
        /*!
         * \param[in] fn Filename
         */
        named_ios(const string& fn) : named_file(fn) {
            constraints::derived_from<stream_type, std::ios>();
        }
    };

    /*!
     * \brief Named version of \c std::ofstream.
     * \sa named_ifstream
     * \sa named_fstream
     */
    class named_ofstream : public ofstream, public named_ios<ofstream> {
    public:
        /*!
         * \param[in] fn Filename
         * \param[in] mode Open mode
         */
        named_ofstream(const string & fn, ios_base::openmode mode = ios_base::out)
            : ofstream(fn.c_str(), mode), named_ios<ofstream>(fn)
        {}
    };

    /*!
     * \brief Named version of \c std::ifstream
     */
    class named_ifstream : public ifstream, public named_ios<ifstream> {
    public:
        /*!
         * \param[in] fn Filename
         * \param[in] mode Open mode
         */
        named_ifstream(const string & fn, ios_base::openmode mode = ios_base::in)
            : ifstream(fn.c_str(), mode), named_ios<ifstream>(fn)
        {}
    };

    /*!
     * \brief Named version of \c std::fstream
     */
    class named_fstream : public fstream, public named_ios<fstream> {
    public:
        /*!
         * \param[in] fn Filename
         * \param[in] mode Open mode
         */
        named_fstream(const string & fn, ios_base::openmode mode = ios_base::in|ios_base::out)
            : fstream(fn.c_str(), mode), named_ios<fstream>(fn)
        {}
    };

} // namespace file

} // namespace net_outlyer

#endif // ONET_NAMED_FSTREAM_HXX
// vim:set ts=4 et ai:

