/*
 * Copyright (C) 2006, 2007, 2009 Toni Corvera
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

// $Id: str_utils.cc 1101 2009-04-05 17:33:40Z  $

#include "str_utils.h"

#include <algorithm> // std::transform
#include <cctype>    // std::toupper, std::tolower
#include <cassert>

#include <iostream>

namespace net_outlyer {

namespace str {

using std::string;
using std::transform;
using std::vector;
using std::max;

// [C++] local visibility
namespace {
    /*!
     * \brief Creates a copy of the input string by applying a transformation
     * \param[in] s Input string
     * \param[in] callback Transformation function to apply
     * \return A copy of \em s after applying \em callback
     * \remarks \em callback will be applied to each input character
     */
    inline
    string copy_transform(const string & s, int(*callback)(int)) {
        string cp(s);
        transform(s.begin(), s.end(), cp.begin(), callback);
        return cp;
    }
}

string to_upper(const string & s) {
    return copy_transform(s, std::toupper);
}

string to_lower(const string & s) {
    return copy_transform(s, std::tolower);
}

string to_hex(int n) {
    return to_string<int>(n, std::hex);
}

string to_hex(int n, unsigned int minbytes, bool zeroex) {
    return ( zeroex ? "0x" : "" ) + to_string(n, std::hex, minbytes*2);
}

string pad(const string & s, size_t pad_to, const string & pad, const pad_direction & p) {
    if (s.length() >= pad_to) return s;

    string cp;
    cp.reserve(pad_to);

    const size_t diff = pad_to - s.length();

    assert( diff > 0 );

    switch (p) {
        case PAD_RIGHT:
        case PAD_LEFT:
        {
            string padding(pad);
            while (padding.length() < diff) {
                padding.append(pad);
            }

            if (PAD_RIGHT == p) {
                cp = s + padding.substr(0, diff);
            }
            else {
                assert( PAD_LEFT == p );
                cp = padding.substr(0, diff) + s;
            }
        }
        break;
        case PAD_BOTH:
        {
            const size_t llen = diff / 2;
            const size_t rlen = diff - llen;

            string padding(pad);
            while (padding.length() < max<size_t>(llen, rlen)) {
                padding.append(pad);
            }

            cp = padding.substr(0, llen) + s + padding.substr(0, rlen);
        }
        break;
        default:
            assert( false );
    } // switch

    return cp;

} // pad [string]

string pad(const string & s, size_t pad_to, char pad, const pad_direction & p) {
    if (s.length() >= pad_to) return s;

    const size_t diff = pad_to - s.length();
    assert( diff > 0 );

    string cp;
    cp.reserve(pad_to);

    switch (p) {
        case PAD_RIGHT:
        case PAD_LEFT:
        {
            const string padding(diff, pad);

            if (PAD_RIGHT == p) {
                cp = s + padding;
            }
            else {
                assert( PAD_LEFT == p );
                cp = padding + s;
            }
        }
        break;
        case PAD_BOTH:
        {
            const size_t llen = diff / 2;
            const size_t rlen = diff - llen;

            const string pl(llen, pad);
            const string pr(rlen, pad);

            cp = pl + s + pr;
        }
        break;
        default:
            assert( false );
    } // switch

    return cp;
} // pad [char]

bool ends_in(const string & s, const string & end) {
    assert( end.length() > 0 );

    if (end.length() > s.length()) return false;

    return s.substr(s.length() - end.length()) == end;
}


} // namespace str

} // namespace net_outlyer

// vim:set ts=4 et ai:
