/*
 * Copyright (C) 2006, 2007, 2009 Toni Corvera
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

// $Id: str_utils.h 1344 2009-06-22 01:11:09Z toni $

/*!
 * \file
 * \brief String utility functions.
 * \remarks Error handling is not always graceful
 *
 * History:
 * \li \c r8 2009-04-27: Added constraints
 * \li \c r7 2009-04-05:
 *                       - Added filling to \c to_string: Changed
 *                         \code to_string(T,base_t) \endcode to
 *                         \code to_string(T,base_t,int minwidth) \endcode
 *                       - Added \c char padding function
 *                         \code pad(string,size_t,char,pad_direction) \endcode
 *                       - Changed license to Two-clause BSD
 * \li \c r6 2007-05-28: Expanded doxygen docs
 * \li \c r5 2007-05-27:
 *                       - Moved less common/generic functions to a separate module
 *                         (removed: \c auto_cstr, \c ieee_size, \c percentage_bar,
 *                         \c replace_all, \c split)
 *                       - Shorter namespace (\c str_utils -> \c str)
 * \li \c r4 2007-05-17: Added GPL license text
 * \li \c r3 2007-05-07: Added \code to_hex(int,unsigned int,bool) \endcode
 * \li \c r2 2007-03-12: Doxygen documentation
 * \li \c r1 2007-03-06: First (recorded) version
 */
#ifndef ONET_STR_UTILS_H
#define ONET_STR_UTILS_H

#include <string>
#include <sstream>
#include <vector>
#include <cassert>
#include <iomanip> // setfill, setw
#include "constraints.hxx"

namespace net_outlyer {

/*!
 * \brief This namespace contains miscellaneous string utilities
 * \todo TODO: Optimise where appropiate
 */
namespace str {

/*!
 * \brief alias for std::ios_base callbacks
 * \remarks base_t is the same as std::ios_base & (*fn)(std::ios_base &)
 * \see to_string(const T&, base_t)
 */
typedef std::ios_base & (base_t)(std::ios_base &);

/*!
 * \brief Converts a given value to a string
 * \param[in] v Value to convert
 * \param[in] base Numeric base to use. Only meaningful for scalars (decimal by default).
 * \param[in] minwidth minimal width
 * \pre \c v converts to \c std::string or there's an associated stream output operator
 * \see http://www.codeguru.com/forum/showthread.php?t=231056 for the original
 * \remarks T should be able to print to a stream
 * \see to_hex(int)
 */
template<typename T> std::string to_string(const T & v,
                                           base_t base = std::dec,
                                           int minwidth = 0)
{
    constraints::is_printable<T>();
    std::ostringstream ss;
    ss << std::setfill('0') << std::setw(minwidth) << base << v;
    return ss.str();
}

//! Shorthand version of to_string(const T&, base_t)
inline std::string to_string(int n) { return to_string<int>(n); }
//! Shorthand version of to_string(const T&, base_t)
inline std::string to_string(unsigned int n) { return to_string<unsigned int>(n); }

/*!
 * \brief Returns the hex representation of a scalar number
 * \see to_string(const T&, base_t)
 */
std::string to_hex(int);

/*!
 * \brief Like to_hex(int) but pretty printing
 * \param[in] v Input value
 * \param[in] minbytes Minimum number of bytes to print (zeros will be added to
 *                     pad) note that 1 means *two* characters
 * \param[in] prepend_zero_x Prepend 0x
 */
std::string to_hex(int v, unsigned int minbytes, bool prepend_zero_x = false);

/*!
 * \brief Returns an uppercase copy of the string
 * \param[in] s Input string
 * \return A new string, same as input string but with uppercase characters
 */
std::string to_upper(const std::string & s);

/*!
 * \brief Returns a lowercase copy of the string
 */
std::string to_lower(const std::string &);

/*!
 * \brief Defines the direction of the padding
 * \see pad()
 */
enum pad_direction {
    PAD_LEFT, /*!< \brief Add padding on the left */
    PAD_RIGHT,    /*!< \brief Add padding on the right */

    /*!
     * \brief Add padding by both sides
     * \note When using \em PAD_BOTH, if padding is odd, the right padding
     *       will be bigger than the right padding
     */
    PAD_BOTH
};

/*!
 * \brief Returns a copy of s padded with padding
 * \param[in] s Input string
 * \param[in] pad_to Minimum number of characters
 * \param[in] padding String to use as padding
 * \param[in] pad_dir Direction of the padding (where will the padding be
 * added)
 * \return A new string with the padded input string
 */
std::string pad(const std::string & s,
                size_t pad_to,
                const std::string & padding = " ",
                const pad_direction & pad_dir = PAD_RIGHT);

/*!
 * \brief Returns a copy of s padded with padding
 * \param[in] s Input string
 * \param[in] pad_to Minimum number of characters
 * \param[in] padding character to use as padding
 * \param[in] pad_dir Direction of the padding (where will the padding be
 * added)
 * \return A new string with the padded input string
 */
std::string pad(const std::string & s,
                size_t pad_to,
                char padding,
                const pad_direction & pad_dir = PAD_RIGHT);

/*!
 * \brief Checks if a string's ending matches a second given string
 * \param[in] s string to check
 * \param[in] end ending to look for
 * \return true if \em s ends in \em ending
 */
bool ends_in(const std::string & s, const std::string & end);

} // namespace str

} // namespace net_outlyer

#endif // ONET_STR_UTILS_H
// vim:set ts=4 et ai:
